# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from math import ceil
from typing import Optional, Union, List, Dict

from pymatgen.core import Composition, Structure, Element
from pymatgen.io.vasp.sets import Potcar

from vise.analyzer.band_edge_properties import is_band_gap
from vise.defaults import defaults
from vise.input_set.datasets.dataset_util import num_bands, calc_kpar, LDAU
from vise.input_set.fft_grids import vasp_grid
from vise.input_set.task import Task
from vise.input_set.xc import Xc
from vise.util.logger import get_logger

logger = get_logger(__name__)


class IncarSettingsGenerator:
    def __init__(
            self,
            structure: Structure,
            symbol_list: list,
            num_kpts: int,
            num_kpt_multiplication_factor: int,
            potcar: Potcar,
            task: Task,
            xc: Xc,
            dos_step_size: float = defaults.dos_step_size,
            charge: float = 0.0,
            band_gap: Optional[float] = None,
            # [vbm, cbm] in absolute eV
            vbm_cbm: Optional[List[float]] = None,
            exchange_ratio: float = 0.25,
            set_hubbard_u: Optional[bool] = None,
            auto_kpar: bool = True,
            cutoff_energy: Optional[float] = None,
            is_magnetization: bool = False,
            num_cores_for_kpar: int = defaults.default_num_cores,
            unused_core_ratio_threshold: float = defaults.unused_core_ratio_threshold,
            str_opt_encut_multi_factor: float = defaults.str_opt_encut_factor,
            multiples_for_grids: Optional[List[int]] = None,
            set_spin_orbit: bool = False):

        self._composition: Composition = structure.composition
        self._lattice = structure.lattice
        self._symbol_list = symbol_list
        self._num_kpt_multiplication_factor = num_kpt_multiplication_factor
        self._potcar = potcar
        self._num_kpts = num_kpts
        self._task = task
        self._xc = xc
        self._dos_step_size = dos_step_size
        self._charge = charge
        self._band_gap = band_gap
        self._vbm_cbm = vbm_cbm
        self._exchange_ratio = exchange_ratio
        self._auto_kpar = auto_kpar
        self._cutoff_energy = cutoff_energy
        self._is_magnetization = is_magnetization
        self._num_cores_for_kpar = num_cores_for_kpar
        self._unused_core_ratio_threshold = unused_core_ratio_threshold
        self._str_opt_encut_multi_factor = str_opt_encut_multi_factor
        self._multiples_for_grids = multiples_for_grids
        self._set_spin_orbit = set_spin_orbit

        self._incar_settings = {}
        self._set_incar_settings(set_hubbard_u)

    def _set_incar_settings(self, set_hubbard_u):
        self._set_default_settings()
        self._set_task_related_settings()
        self._set_xc_related_settings()
        self._set_options_related_settings()
        self._set_spin_orbit_related_settings()
        if self._task.is_spectrum_task:
            self._set_spectrum_related_settings()
        if self._task.is_dielectric:
            self._set_dielectric_related_settings()
        if self._xc.is_hybrid_functional:
            self._set_hybrid_func_related_settings()
        if self._need_hubbard_u(set_hubbard_u):
            self._set_hubbard_u_related_settings()
        if self._auto_kpar:
            self._set_kpar()
        self._remove_incar_settings_with_none_values()

    def _set_default_settings(self):
        self._incar_settings.update({
            "LASPH": True,
            "NELM": 100,
            "SIGMA": 0.1,
            "LCHARG": False,
        })

    def _set_task_related_settings(self):
        t = TaskIncarSettings(self._task)
        self._incar_settings.update({
            "PREC": t.prec,
            "LREAL": t.lreal,
            "EDIFF": t.ediff,
            "ADDGRID": t.addgrid_optional,
            "ISIF": t.isif,
            "IBRION": t.ibrion,
            "EDIFFG": t.ediffg_optional,
            "NSW": t.nsw,
            "POTIM": t.potim_optional,
            "LORBIT": t.lorbit,
        })

    def _set_xc_related_settings(self):
        x = XcIncarSettings(self._xc)
        self._incar_settings.update({
            "ALGO": x.algo,
            "LWAVE": x.lwave,
            "GGA": x.gga_optional,
            "METAGGA": x.metagga_optional,
        })

    def _set_options_related_settings(self):
        self._incar_settings.update({
            "ENCUT": self._encut,
            "ISMEAR": self._ismear,
            "ISPIN": self._ispin,
            "NBANDS": self._nbands,
            "NELECT": self._nelect,
        })
        if self._multiples_for_grids:
            grids = [vasp_grid(self._incar_settings["ENCUT"],
                               length,
                               self._incar_settings["PREC"])
                     for length in self._lattice.abc]
            m = self._multiples_for_grids
            ngx, ngy, ngz = (ceil(grids[i] / m[i]) * m[i] for i in range(3))
            self._incar_settings.update({"NGX": ngx})
            self._incar_settings.update({"NGY": ngy})
            self._incar_settings.update({"NGZ": ngz})

    def _set_spin_orbit_related_settings(self):
        if self._set_spin_orbit:
            self._incar_settings["LSORBIT"] = True

    def _need_hubbard_u(self, set_hubbard_u):
        if isinstance(set_hubbard_u, bool):
            return set_hubbard_u
        return False

    @property
    def _encut(self) -> float:
        if self._cutoff_energy:
            return self._cutoff_energy
        max_enmax = max([p.enmax for p in self._potcar])

        if self._task.is_lattice_relaxed:
            max_enmax *= self._str_opt_encut_multi_factor
        return round(max_enmax, 3)

    @property
    def _ismear(self) -> int:
        # Tetrahedron method fails for irrep. NKPT<4 in vasp.
        if self._task is Task.band:
            return 0
        elif is_band_gap(self._band_gap, self._vbm_cbm, show_info=True) \
                and self._num_kpts >= 4:
            if self._task in (Task.dos, Task.dielectric_function):
                # tested -4 and -5 show the same results for spectra.
                return -4
            else:
                return -5
        return 0

    @property
    def _ispin(self) -> Optional[int]:
        if self._is_magnetization or self._task.need_spin:
            return 2
        return

    @property
    def _nbands(self) -> Union[int, None]:
        if self._task.is_plot_task:
            return num_bands(self._composition, self._potcar,
                             self._set_spin_orbit)
        return

    @property
    def _nelect(self) -> Union[float, None]:
        if self._charge:
            comp = self._composition.element_composition
            nelect = sum([comp[pt.element] * pt.ZVAL for pt in self._potcar])
            return nelect - self._charge
        return

    def _set_spectrum_related_settings(self) -> None:
        # Need one more step for vasp to remove weired huge value.
        if self._vbm_cbm:
            emin = ceil(self._vbm_cbm[0]) - 15 - self._dos_step_size
            emax = ceil(self._vbm_cbm[1]) + 15
        else:
            emin = -20 - self._dos_step_size
            emax = 20
        self._incar_settings["EMIN"] = emin
        self._incar_settings["EMAX"] = emax
        num_steps = round((emax - emin) / self._dos_step_size)
        self._incar_settings["NEDOS"] = num_steps + 1

    def _set_dielectric_related_settings(self) -> None:
        if self._task == Task.dielectric_dfpt:
            self._incar_settings["LEPSILON"] = True
        elif self._task == Task.dielectric_finite_field:
            self._incar_settings["LCALCEPS"] = True
#            self._incar_settings["IBRION"] = 6
            self._incar_settings["POTIM"] = 0.015
        elif self._task == Task.dielectric_function:
            self._incar_settings["LOPTICS"] = True
            # Since the real part of dielectric function depends on the vasp
            # version, we here adapt CSHIFT=0.0 and apply KK transformation
            # using vise.
            self._incar_settings["CSHIFT"] = 0.0

    def _set_hybrid_func_related_settings(self) -> None:
        self._incar_settings["LHFCALC"] = True
        self._incar_settings["PRECFOCK"] = "Fast"
        self._incar_settings["TIME"] = 0.4
        self._incar_settings["AEXX"] = self._exchange_ratio
        if self._num_kpt_multiplication_factor != 1:
            self._incar_settings["NKRED"] = self._num_kpt_multiplication_factor
        if self._xc is Xc.hse:
            self._incar_settings["HFSCREEN"] = 0.208

    def _set_hubbard_u_related_settings(self) -> None:
        ldau = LDAU(self._symbol_list)
        if ldau.is_ldau_needed:
            self._incar_settings["LDAU"] = True
            self._incar_settings["LDAUTYPE"] = 2
            self._incar_settings["LMAXMIX"] = ldau.lmaxmix
            self._incar_settings["LDAUPRINT"] = 1
            self._incar_settings["LDAUU"] = ldau.ldauu
            self._incar_settings["LDAUL"] = ldau.ldaul

    def _set_kpar(self) -> None:
        if self._kpar_incompatible():
            return
        self._incar_settings["KPAR"] = \
            calc_kpar(self._num_kpts,
                      self._num_cores_for_kpar,
                      self._unused_core_ratio_threshold)
        # now switch off NPAR
        # self._settings["NPAR"] = npar

    def _kpar_incompatible(self):
        if self._incar_settings.get("LELF", False):
            logger.warning("Because KPAR cannot be set with LELF=True,"
                           "KPAR is switched off.")
            return True

    def _remove_incar_settings_with_none_values(self) -> None:
        for tag_name, value in self._incar_settings.copy().items():
            if value is None:
                self._incar_settings.pop(tag_name)

    @property
    def incar_settings(self) -> Dict[str, Union[str, bool, int, float]]:
        return self._incar_settings


class TaskIncarSettings:
    def __init__(self, task: Task):
        self._task = task

    @property
    def isif(self):
        if self._task.is_lattice_relaxed:
            return 3
        elif (self._task.is_atom_relaxed_lattice_fixed
              or self._task is Task.phonon_force):
            return 2
        elif self._task in (Task.band, Task.dos) or self._task.is_dielectric:
            return 0
        else:
            raise NotImplementedError

    # During dielectric_dfpt calculations, EDIFF is tightened automatically.
    @property
    def ediff(self):
        if self._task.is_tight_calc:
            return 1e-8
        elif self._task in (Task.structure_opt, Task.cluster_opt):
            return 1e-7
        elif self._task in (Task.dielectric_dfpt, Task.dielectric_finite_field):
            return 1e-6
        elif self._task in (Task.dielectric_function,
                            Task.band,
                            Task.dos,
                            Task.defect,
                            Task.defect_2d):
            return 1e-5
        elif self._task in (Task.structure_opt_rough,):
            return 1e-4
        else:
            raise NotImplementedError

    @property
    def ediffg_optional(self):
        if self._task.is_atom_relaxed:
            if self._task is Task.structure_opt_tight:
                return -0.001
            elif self._task in (Task.structure_opt, Task.cluster_opt):
                return -0.005
            elif self._task in (Task.defect, Task.defect_2d):
                return -0.03
            elif self._task is Task.structure_opt_rough:
                return -0.2
            else:
                raise NotImplementedError
        else:
            return

    @property
    def ibrion(self):
        return 8 if self._task is Task.dielectric_dfpt else 2

    @property
    def lorbit(self):
        if self._task is Task.phonon_force:
            return None
        return 10 if self._task is not Task.dos else 11

    @property
    def lreal(self):
        return "Auto" if self._task in (Task.defect, Task.defect_2d) else False

    @property
    def prec(self):
        return "Accurate" if self._task.is_tight_calc else "Normal"

    @property
    def nsw(self):
        return 20 if self._task.is_atom_relaxed else 1

    @property
    def potim_optional(self):
        if self._task is Task.structure_opt_rough:
            return 0.1
        return

    @property
    def addgrid_optional(self):
        if self._task.is_tight_calc:
            return True
        return


class XcIncarSettings:
    def __init__(self, xc: Xc):
        self._xc = xc

    @property
    def algo(self):
        return "Damped" if self._xc.is_hybrid_functional else "Normal"

    @property
    def lwave(self):
        return True if self._xc.is_hybrid_functional else False

    @property
    def gga_optional(self):
        return "PS" if self._xc is Xc.pbesol else None

    @property
    def metagga_optional(self):
        return "SCAN" if self._xc is Xc.scan else None


















# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np
import pytest

from monty.serialization import loadfn
from vise.analyzer.effective_mass import EffectiveMass, eigvals_and_vecs, \
    lowest_eigval_and_vecs


@pytest.fixture
def em():
    return EffectiveMass(p=[[[1.0, 0, 0], [0, 1.0, 0], [0, 0, 1.0]]],
                         n=[[[2.0, 0, 0], [0, 2.0, 0], [0, 0, 2.0]]],
                         temperature=300,
                         concentrations=[10**18])


def test_str(em):
    expected = """temperature: 300
------------------------------
concentration: 1e+18
p:
-  -  -
1  0  0
0  1  0
0  0  1
-  -  -
n:
-  -  -
2  0  0
0  2  0
0  0  2
-  -  -"""
    assert str(em) == expected


def test_effective_mass_json_file_mixin(em, tmpdir):
    tmpdir.chdir()
    em.to_json_file()
    actual = loadfn("effective_mass.json")
    assert actual.p == em.p


def test_effective_mass(em):
    actual = em.effective_mass("p", 10**18)
    assert actual == em.p[0]
    actual = em.average_mass("p", 10**18)
    assert actual == 1.0
    actual = em.minimum_mass("p", 10**18)
    assert actual == 1.0


def test_lowest_eigval_and_vec(em):
    matrix = np.array([[1.0, 0.2, 0.1], [0.2, 2.0, 0.3], [0.1, 0.3, 3.0]])
    actual_vals, actual_vecs = eigvals_and_vecs(matrix)
    expected_vals = [0.961, 1.945, 3.094]
    vec = np.array([-46.194, 8.6, 1.0])
    norm = np.linalg.norm(vec, ord=2)
    expected_vec0 = vec / norm
    np.testing.assert_array_almost_equal(actual_vals, expected_vals, decimal=3)
    np.testing.assert_array_almost_equal(actual_vecs[0], expected_vec0)

    actual_lowest_eigval, actual_eigvecs = lowest_eigval_and_vecs(matrix)
    expected_lowest_eigval = 0.9606010675519356
    expected_eigvecs = np.array([[-0.98,  0.18,  0.02]])
    assert actual_lowest_eigval == expected_lowest_eigval
    np.testing.assert_array_almost_equal(actual_eigvecs, expected_eigvecs)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest
from pymatgen.core import Structure, Lattice
from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp import Vasprun, Outcar, Procar
from vise.analyzer.band_edge_properties import BandEdge

from vise.analyzer.vasp.band_edge_properties import eigenvalues_from_vasprun, \
    VaspBandEdgeProperties, edge_orbital_contributions
from vise.tests.analyzer.test_band_edge_properties import parent_dir


def test_band_edge_properties_from_vasp(test_data_files):
    vasprun_file = str(test_data_files / "MnO_uniform_vasprun.xml")
    vasprun = Vasprun(vasprun_file)
    outcar_file = str(test_data_files / "MnO_uniform_OUTCAR")
    outcar = Outcar(outcar_file)
    band_edge = VaspBandEdgeProperties(vasprun, outcar)
    assert pytest.approx(band_edge.band_gap) == 0.4702


def test_band_edge_properties_from_vasp_non_mag(test_data_files):
    vasprun_file = str(test_data_files / "MgO_band_vasprun.xml")
    vasprun = Vasprun(vasprun_file)
    outcar_file = str(test_data_files / "MgO_band_OUTCAR")
    outcar = Outcar(outcar_file)
    band_edge = VaspBandEdgeProperties(vasprun, outcar)

    assert pytest.approx(band_edge.band_gap) == 4.6597


def test_band_edge_properties_from_vasp_non_collinear(test_data_files):
    vasprun_file = str(test_data_files / "MgSe_band_noncollinear_vasprun.xml")
    vasprun = Vasprun(vasprun_file)
    outcar_file = str(test_data_files / "MgSe_band_noncollinear_OUTCAR")
    outcar = Outcar(outcar_file)
    band_edge = VaspBandEdgeProperties(vasprun, outcar)

    assert pytest.approx(band_edge.band_gap) == 1.3199


def test_band_edge_properties_orbital_contributions(test_data_files):
    procar = Procar(test_data_files / "MgO_band_PROCAR")
    structure = Structure(
        Lattice.cubic(1), species=["Mg", "O"], coords=[[0]*3, [0.5]*3])
    vbm_info = BandEdge.from_dict({'energy': 3.0663, 'spin': 1, 'band_index': 3,
                                   'kpoint_index': 0,
                                   'kpoint_coords': [0.0, 0.0, 0.0]})
    cbm_info = BandEdge.from_dict({'energy': 7.726, 'spin': 1, 'band_index': 4,
                                   'kpoint_index': 0,
                                   'kpoint_coords': [0.0, 0.0, 0.0]})
    vbm, cbm = edge_orbital_contributions(procar, structure, vbm_info, cbm_info)
    assert vbm["O"]["p"] == 0.753
    assert cbm["Mg"]["s"] == 0.238


def test_eigenvalues_from_vasprun(test_data_files):
    vasprun_file = str(test_data_files / "MnO_uniform_vasprun.xml")
    vasprun = Vasprun(vasprun_file)
    assert eigenvalues_from_vasprun(vasprun)[Spin.up][1, 2] == -0.1397





# -*- coding: utf-8 -*-
#  Copyright (c) 2021. Distributed under the terms of the MIT License.
from numpy import prod
from typing import List

import numpy as np
from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp import Poscar, VolumetricData, Chgcar
from vise.util.logger import get_logger

logger = get_logger(__name__)

_minor = 1e-3
default_border_fractions = [0.1, 0.5, 0.8]


def make_spin_charges(chgcar: Chgcar) -> List[Chgcar]:
    result = [Chgcar(chgcar.structure, {"total": chgcar.spin_data[Spin.up]})]
    if "diff" in chgcar.data:
        result.append(
            Chgcar(chgcar.structure, {"total": chgcar.spin_data[Spin.down]}))
    return result


def light_weight_vol_text(volumetric_data: VolumetricData,
                          border_fractions: List[float] = None):
    data = np.zeros(prod(volumetric_data.dim), dtype=int)
    normalized_values = (volumetric_data.data["total"]
                         / np.max(volumetric_data.data["total"])) + _minor

    border_fractions = border_fractions or default_border_fractions
    for border in border_fractions:
        # transpose needed as vasp is based on column measure (Fortran)
        data += (normalized_values > border).T.flatten()

    lines = [Poscar(volumetric_data.structure).get_str(),
             " ".join([str(d) for d in volumetric_data.dim]),
             " ".join(data.astype(str))]
    return "\n".join(lines)
    # filename.write_text("\n".join(lines))





# -*- coding: utf-8 -*-
#  Copyright (c) 2021. Distributed under the terms of the MIT License.
from pymatgen.io.vasp import Potcar
from vise.input_set.potcar_generator import generate_potcar
from vise.input_set.xc import Xc
from vise.util.valence_orbitals_from_potcar import valence_orbitals_from_potcar


def test_valence_orbitals_from_potcar():
    potcar_single = generate_potcar(["Mg"], xc=Xc.pbe)[0]
    assert valence_orbitals_from_potcar(potcar_single) == "(3s)2"



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import os
from argparse import Namespace
from copy import deepcopy
from pathlib import Path

import pytest
from pymatgen.core import Structure, Element
from vise.analyzer.atom_grouping_type import AtomGroupingType
from vise.analyzer.dielectric_function import DieleFuncData
from vise.analyzer.plot_diele_func_data import DieleFuncPlotType
from vise.cli.main_functions import get_poscar_from_mp, VaspSet, plot_band, \
    plot_dos, band_edge_properties, plot_diele_func, \
    structure_info, get_most_stable_mp_id_from_formula, \
    get_most_stable_mp_id_from_formula_w_new_mprester
from vise.defaults import defaults
from vise.input_set.kpoints_mode import KpointsMode
from vise.input_set.task import Task
from vise.input_set.xc import Xc

default_option_args = {"poscar": Path("POSCAR"),
                       "task": Task.structure_opt,
                       "xc": Xc.pbe,
                       "kpt_density": 1.0,
                       "overridden_potcar": ["Mn_pv"]}

default_args = deepcopy(default_option_args)
default_args.update({"dirs": [Path.cwd()],
                     "user_incar_settings": None,
                     "prev_dir": None,
                     "options": None,
                     "file_transfer_type": None,
                     "uniform_kpt_mode": False,
                     "vasprun": Path("vasprun.xml"),
                     "outcar": Path("outcar.xml"),
                     })


def test_structure_info(mocker):
    args = Namespace(poscar=Path("POSCAR"), symprec=0.1, angle_tolerance=5,
                     show_conventional=False, show_primitive=False)
    lattice = [[10.0,  0.0,  0.0], [0.0, 10.0,  0.0], [-2.0,  0.0, 10.0]]
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.0]]

    mock = mocker.patch("vise.cli.main_functions.Structure")
    mock.from_file.return_value = Structure(lattice=lattice, species=["H"] * 2,
                                            coords=coords)
    structure_info(args)
    args = Namespace(poscar=Path("POSCAR"), symprec=0.1, angle_tolerance=5,
                     show_conventional=True, show_primitive=False)
    structure_info(args)
    args = Namespace(poscar=Path("POSCAR"), symprec=0.1, angle_tolerance=5,
                     show_conventional=True, show_primitive=True)
    structure_info(args)

#
# def test_get_most_stable_mp_id_from_formula():
#     actual = get_most_stable_mp_id_from_formula_w_new_mprester("MgO")
#     assert actual == "mp-1265"
#    actual = get_most_stable_mp_id_from_formula("MgO")
#    assert actual == "mp-1265"

#
# def test_get_most_stable_mp_id_from_formula_not_present_formula():
#     actual = get_most_stable_mp_id_from_formula_w_new_mprester("MgO100")
#     assert actual is None


mg = """Mg1
1.0
   2.8949120000000002    0.0000000000000000   -1.0235060000000000
  -1.4474560000000001    2.5070679999999999   -1.0235060000000000
   0.0000000000000000    0.0000000000000000    3.0705179999999999
Mg
1
direct
   0.0000000000000000    0.0000000000000000    0.0000000000000000 Mg"""
mg_110_structure = Structure.from_str(mg, fmt="poscar")


def test_get_poscar_from_mp(tmpdir):
    args = Namespace(mpid="mp-110", formula="He")
    tmpdir.chdir()
    print(tmpdir)
    get_poscar_from_mp(args)
    assert Structure.from_file("POSCAR") == mg_110_structure
    assert Path("prior_info.yaml").read_text() == """band_gap: 0.0
data_source: new MPRester mp-110
total_magnetization: 0.0
"""
    os.remove("prior_info.yaml")

mg_153 = """Mg2
1.0
   3.1720237199999999    0.0000000000000000   -0.0000000000000000
  -1.5860113600000001    2.7470524099999998    0.0000000000000000
  -0.0000000000000000   -0.0000000000000000    5.1412049000000000
Mg
2
direct
   0.3333330000000000    0.6666670000000000    0.2500000000000000 Mg
   0.6666670000000000    0.3333330000000000    0.7500000000000000 Mg
"""

mg_153_structure = Structure.from_str(mg_153, fmt="poscar")


def test_get_poscar_from_mp_by_formula(tmpdir):
    args = Namespace(mpid=None, formula="Mg")
    print(tmpdir)
    tmpdir.chdir()
    get_poscar_from_mp(args)
    assert Structure.from_file("POSCAR") == mg_153_structure
    assert Path("prior_info.yaml").read_text() == """band_gap: 0.0
data_source: new MPRester mp-153
total_magnetization: 0.0
"""
    # Need to remove file to avoid the side effect for other unittests.
    os.remove("prior_info.yaml")


def test_get_poscar_from_mp_wo_arguments():
    args = Namespace(mpid=None, formula=None)
    with pytest.raises(ValueError):
        get_poscar_from_mp(args)


test_data = [
    ({}, {}, {}, {}),

    ({"user_incar_settings": ["ISPIN", "2"]},
     {"ISPIN": 2},
     {},
     {}),

    ({"options": ["only_even_num_kpts", "True"]},
     {},
     {"only_even_num_kpts": True},
     {}),

    ({"uniform_kpt_mode": True},
     {},
     {"kpt_mode": KpointsMode.uniform},
     {}),

    ({"prev_dir": Path("a"), "file_transfer_type": ["file", "c"]},
     {},
     {},
     {"x": "y"}),

    ({"options": ["kpt_density", "3.5"], "kpt_density": 2.5},
     {},
     {"kpt_density": 2.5},
     {})
]


# Be careful about literal; all the words are connected (see position of commas)
@pytest.mark.parametrize("modified_settings,"
                         "overridden_incar_settings,"
                         "overridden_options_args,"
                         "prior_info", test_data)
def test_user_incar_settings(mocker,
                             modified_settings,
                             overridden_incar_settings,
                             overridden_options_args,
                             prior_info):
    args = deepcopy(default_args)
    args.update(modified_settings)

    structure = mocker.patch("vise.cli.main_functions.Structure")
    prior_info_mock = mocker.patch(
        "vise.cli.main_functions.prior_info_from_calc_dir")
    options = mocker.patch("vise.cli.main_functions.CategorizedInputOptions")
    vif = mocker.patch("vise.cli.main_functions.VaspInputFiles")

    prior_info_mock.return_value.input_options_kwargs = prior_info

    name_space = Namespace(**args)
    VaspSet(name_space)

    option_args = deepcopy(default_option_args)
    option_args.update(overridden_options_args)
    option_args.update(prior_info)
    option_args["overridden_potcar"] = {"Mn": "Mn_pv"}
    option_args.pop("poscar")
    option_args["structure"] = structure.from_file.return_value

    options.assert_called_once_with(**option_args)

    incar_settings = defaults.user_incar_settings
    incar_settings.update(overridden_incar_settings)

    vif.assert_called_once_with(options.return_value, incar_settings)


# def test_user_incar_settings_with_kpt_density_in_options():
#     args = deepcopy(default_args)
#     args.update(dict(kpt_density=2.5, options={"kpt_density": 3.5}))
#     name_space = Namespace(**args)
#     VaspSet(name_space)


def test_plot_band(tmpdir, test_data_files):
    tmpdir.chdir()
    args = Namespace(vasprun=test_data_files / "H_band_vasprun.xml",
                     kpoints_filename=str(test_data_files / "H_band_KPOINTS"),
                     y_range=[-10, 10],
                     wavecar_filename=None,
                     poscar=str(test_data_files / "H_band_POSCAR"),
                     plotly=False,
                     filename="test.pdf")

    plot_band(args)


def test_plot_band_irrep_and_plotly(test_data_files):
    args = Namespace(vasprun=test_data_files / "H_band_vasprun.xml",
                     kpoints_filename=str(test_data_files / "H_band_KPOINTS"),
                     y_range=[-10, 10],
                     wavecar_filename=str(test_data_files / "H_band_WAVECAR"),
                     poscar=str(test_data_files / "H_band_POSCAR"),
                     plotly=True,
                     filename="test.pdf")
    plot_band(args)


def test_plot_dos(tmpdir, test_data_files):
    tmpdir.chdir()  # comment out when one wants to see the figure
    args = Namespace(vasprun=test_data_files / "MnO_uniform_vasprun.xml",
                     outcar=test_data_files / "MnO_uniform_OUTCAR",
                     type=AtomGroupingType.non_equiv_sites,
                     legend=True,
                     base_energy=None,
                     crop_first_value=True,
                     x_range=[-5, 5],
                     y_max_ranges=[10, 5, 7],
                     target=["Mn", "O"],
                     filename="test.pdf",
                     title="test")
    plot_dos(args)


def test_band_edge_info(test_data_files):
    args = Namespace(vasprun=test_data_files / "MnO_uniform_vasprun.xml",
                     outcar=test_data_files / "MnO_uniform_OUTCAR")
    band_edge_properties(args)


def test_plot_diele_func(tmpdir, test_data_files):
    print(tmpdir)
    tmpdir.chdir()
    args = Namespace(vasprun=test_data_files / "MgSe_absorption_vasprun.xml",
                     outcar=test_data_files / "MgSe_absorption_OUTCAR",
                     input_csv_name=None,
                     y_range=[2.0, 8.0],
                     calc_kk=False,
                     ita=0.1,
                     filename=None,
                     directions=["ave"],
                     plot_type=DieleFuncPlotType.absorption_coeff,
                     to_csv=True,
                     title=None)
    plot_diele_func(args)

    args = Namespace(vasprun=test_data_files / "MgSe_absorption_vasprun.xml",
                     outcar=test_data_files / "MgSe_absorption_OUTCAR",
                     input_csv_name="diele_func_data.csv",
                     y_range=None,
                     calc_kk=False,
                     ita=0.1,
                     filename=None,
                     directions=["xx"],
                     plot_type=DieleFuncPlotType.refraction,
                     to_csv=False,
                     title="main func test")
    plot_diele_func(args)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest

from vise.input_set.kpoints_mode import KpointsMode
from vise.input_set.task import Task
from vise.tests.helpers.assertion import assert_msonable


def test_task_str():
    assert "dos" == str(Task.dos)


def test_task_msonable():
    assert_msonable(Task.dos)


def test_task_from_string():
    assert Task.dos == Task.from_string("dos")


def test_task_from_string_raise_error():
    with pytest.raises(AttributeError):
        Task.from_string("not_exist_member")


def test_task_name_list():
    expected = 'structure_opt, structure_opt_rough, structure_opt_tight, ' \
               'cluster_opt, phonon_force, defect, defect_2d, band, dos, ' \
               'dielectric_dfpt, dielectric_finite_field, dielectric_function'
    assert Task.names_string() == expected


def test_is_atom_relaxed_lattice_fixed():
    assert Task.structure_opt_tight.is_atom_relaxed_lattice_fixed is False
    assert Task.cluster_opt.is_atom_relaxed_lattice_fixed is True
    assert Task.band.is_atom_relaxed_lattice_fixed is False


def test_is_lattice_relaxed():
    assert Task.structure_opt.is_lattice_relaxed is True
    assert Task.structure_opt_rough.is_lattice_relaxed is True
    assert Task.band.is_lattice_relaxed is False


def test_is_atom_relaxed():
    assert Task.structure_opt_tight.is_atom_relaxed is True
    assert Task.cluster_opt.is_atom_relaxed is True
    assert Task.band.is_atom_relaxed is False


def test_is_dielectric():
    assert Task.dielectric_function.is_dielectric is True
    assert Task.band.is_dielectric is False


def test_is_tight_calc():
    assert Task.structure_opt_tight.is_tight_calc is True
    assert Task.band.is_tight_calc is False


def test_is_plot_task():
    assert Task.band.is_plot_task is True
    assert Task.structure_opt.is_plot_task is False


def test_is_spectrum_task():
    assert Task.dos.is_spectrum_task is True
    assert Task.band.is_spectrum_task is False


def test_default_kpt_factor():
    assert Task.band.default_kpt_factor == 1
    assert Task.dos.default_kpt_factor == 2
    assert Task.dielectric_dfpt.default_kpt_factor == 2
    assert Task.dielectric_finite_field.default_kpt_factor == 2
    assert Task.dielectric_function.default_kpt_factor == 3


def test_default_kpt_mode():
    assert Task.band.default_kpt_mode == KpointsMode.band
    assert Task.defect.default_kpt_mode == KpointsMode.uniform
    assert Task.dos.default_kpt_mode == KpointsMode.primitive


def test_requisite_num_kpt_list():
    assert Task.cluster_opt.requisite_num_kpt_list == [1, 1, 1]
    assert Task.band.requisite_num_kpt_list is None


def test_requisite_only_even_num_kpts():
    assert Task.cluster_opt.requisite_only_even_num_kpts is False
    assert Task.band.requisite_only_even_num_kpts is True
    assert Task.dos.requisite_only_even_num_kpts is None


def test_requisite_kpt_shift():
    assert Task.cluster_opt.requisite_gamma_centered is True
    assert Task.dos.requisite_gamma_centered is True
    assert Task.band.requisite_gamma_centered is None


def test_need_spin():
    assert Task.defect.need_spin is True
    assert Task.band.need_spin is False


def test_fine_to_inherit_structure_from_prev():
    assert Task.structure_opt.fine_to_inherit_structure_from_prev is True
    assert Task.phonon_force.fine_to_inherit_structure_from_prev is False




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pathlib import Path

import pytest
from pymatgen.core.structure import Structure

from vise import __version__
from vise.defaults import defaults
from vise.input_set.input_options import CategorizedInputOptions
from vise.input_set.task import Task
from vise.input_set.vasp_input_files import VaspInputFiles
from vise.input_set.vise_log import ViseLog
from vise.input_set.xc import Xc
from vise.tests.helpers.assertion import assert_yaml_roundtrip


@pytest.fixture
def vasp_input_files():
    lattice = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]]
    structure = Structure(lattice=lattice, species=["Mn", "O"], coords=coords)
    input_options = CategorizedInputOptions(
        structure=structure, task=Task.structure_opt, xc=Xc.pbe, charge=1,
        set_hubbard_u=True)

    return VaspInputFiles(input_options,
                          overridden_incar_settings={"NSW": 2})


def test_integration(tmpdir, vasp_input_files):
    tmpdir.chdir()
    vasp_input_files.create_input_files(Path.cwd())


def test_vise_log(vasp_input_files):
    expected = ViseLog(version=__version__, task=Task.structure_opt, xc=Xc.pbe,
                       input_options={"charge": 1,
                                      "set_hubbard_u": True,
                                      "kpt_density": defaults.kpoint_density},
                       user_incar_settings={"NSW": 2})
    assert vasp_input_files.vise_log == expected




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from copy import copy
from pathlib import Path
from typing import Optional, Dict, Any

from pymatgen.io.vasp.sets import Poscar, Kpoints, Potcar

from vise import __version__
from vise.input_set.incar import ViseIncar
from vise.input_set.incar_settings_generator import IncarSettingsGenerator
from vise.input_set.input_options import CategorizedInputOptions
from vise.input_set.kpoints import ViseKpoints
from vise.input_set.potcar_generator import generate_potcar
from vise.input_set.structure_kpoints_generator import StructureKpointsGenerator
from vise.input_set.vise_log import ViseLog
from vise.util.logger import get_logger
from vise.util.structure_handler import create_symbol_list

logger = get_logger(__name__)


class VaspInputFiles:
    def __init__(self,
                 input_options: CategorizedInputOptions,
                 overridden_incar_settings: Optional[dict] = None):

        self._version = __version__
        self._input_options = input_options
        self._initial_structure = input_options.initial_structure
        self._generate_structure_kpoints(input_options)
        self._generate_potcar_incar_settings(input_options)
        self._overridden_incar_settings = overridden_incar_settings
        self._incar_settings.update(overridden_incar_settings or {})

    def _generate_structure_kpoints(self, input_options):
        str_kpoints_generator = StructureKpointsGenerator(
            **input_options.structure_kpoints_options)
        str_kpoints_generator.generate_input()

        self._structure = str_kpoints_generator.structure
        self._kpoints = str_kpoints_generator.kpoints
        self._num_kpts = str_kpoints_generator.num_kpts
        self._num_kpt_factor = str_kpoints_generator.num_kpt_factor

    def _generate_potcar_incar_settings(self, input_options):
        symbol_list = create_symbol_list(self._structure)
        self._potcar = generate_potcar(symbol_list=symbol_list,
                                       **input_options.potcar_options)
        incar_settings_generator = \
            IncarSettingsGenerator(self._structure,
                                   symbol_list,
                                   self._num_kpts,
                                   self._num_kpt_factor,
                                   self._potcar,
                                   **input_options.incar_settings_options)
        self._incar_settings = incar_settings_generator.incar_settings

    @property
    def vise_log(self):
        result = copy(self._input_options.vise_log)
        result.user_incar_settings = self._overridden_incar_settings
        return result

    @property
    def initial_structure(self):
        return self._initial_structure

    @property
    def version(self) -> str:
        return self._version

    @property
    def incar(self) -> ViseIncar:
        incar = ViseIncar.from_dict(self._incar_settings)
        return incar

    @property
    def poscar(self) -> Poscar:
        return Poscar(self._structure)

    @property
    def kpoints(self) -> Kpoints:
        return self._kpoints

    @property
    def potcar(self) -> Potcar:
        return self._potcar

    @property
    def input_files(self) -> Dict[str, Any]:
        return {"INCAR": self.incar, "KPOINTS": self.kpoints,
                "POSCAR": self.poscar, "POTCAR": self.potcar}

    def create_input_files(self,
                           dirname: Path,
                           poscar_significant_figures=10) -> None:
        dirname.mkdir(exist_ok=True)
        for filename, obj in self.input_files.items():
            if isinstance(obj, Poscar):
                # Need to show more digit than pymatgen one.
                obj.write_file(dirname / filename,
                               significant_figures=poscar_significant_figures)
            elif isinstance(obj, Kpoints):
                kpoints = ViseKpoints.from_dict(obj.as_dict())
                kpoints.write_file(dirname / filename)
            else:
                obj.write_file(dirname / filename)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import re
from copy import deepcopy
from typing import List
import numpy as np

from pymatgen.electronic_structure.plotter import BSPlotter
from pymatgen.io.vasp import Vasprun
from vise.analyzer.plot_band import BandPlotInfo, BandEnergyInfo, XTicks, \
    BandEdgeForPlot
from vise.analyzer.plot_brillouin_zone import BZPlotInfo
from vise.util.string import latexify


def greek_to_unicode(label: str) -> str:
    # The sequence is important as SIGMA includes GM.
    d = {"GAMMA": "Γ", "SIGMA": "Σ", "GM": "Γ", "DELTA": "Δ", "LAMBDA": "Λ"}
    for k, v in d.items():
        label = label.replace(k, v)
    return label


def italic_to_roman(label: str) -> str:
    return re.sub(r"([A-Z])_([0-9])", r"{\\rm \1}_\2", label)


def _sanitize_label(label: str):
    return italic_to_roman(greek_to_unicode(label))


def _sanitize_labels(labels: str):
    return [_sanitize_label(label) for label in labels]


class BandPlotInfoFromVasp:
    def __init__(self,
                 vasprun: Vasprun,
                 kpoints_filename: str,
                 first_band_plot_name: str = None,
                 wavecar_name: str = None,
                 vasprun2: Vasprun = None,
                 second_band_plot_name: str = None,
                 energy_window: List[float] = None):
        self._composition = vasprun.final_structure.composition
        self.bs = vasprun.get_band_structure(kpoints_filename, line_mode=True)
        self.first_band_plot_name = first_band_plot_name or "1"
        self.wavecar_name = wavecar_name
        self.rlat = vasprun.final_structure.lattice.reciprocal_lattice
        self.kpoints = self.bs.kpoints
        self.energy_window = energy_window

        if vasprun2:
            assert vasprun.final_structure == vasprun2.final_structure
            self.vasprun2 = vasprun2
            self.bs2 = vasprun2.get_band_structure(kpoints_filename,
                                                   line_mode=True)
            self.second_band_plot_name = second_band_plot_name or "2"

        else:
            self.vasprun2, self.bs2, self.second_band_plot_name \
                = None, None, None

    def make_band_plot_info(self) -> BandPlotInfo:
        bs_plotter = BSPlotter(self.bs)
        plot_data = bs_plotter.bs_plot_data(zero_to_efermi=False)
        distances = [list(d) for d in plot_data["distances"]]

        band_info_1 = BandEnergyInfo(
            band_energies=self._remove_spin_key(plot_data),
            band_edge=self._band_edge(self.bs, plot_data),
            fermi_level=self.bs.efermi)
        band_info = {self.first_band_plot_name: band_info_1}

        if self.vasprun2:
            plot_data2 = BSPlotter(self.bs2).bs_plot_data(zero_to_efermi=False)
            band_info[self.second_band_plot_name] = BandEnergyInfo(
                band_energies=self._remove_spin_key(plot_data2),
                band_edge=self._band_edge(self.bs2, plot_data2),
                fermi_level=self.bs2.efermi)

        x = bs_plotter.get_ticks_old()
        x_ticks = XTicks(_sanitize_labels(x["label"]), x["distance"])

        return BandPlotInfo(band_energy_infos=band_info,
                            distances_by_branch=distances,
                            x_ticks=x_ticks,
                            title=self._title)

    def make_bz_plot_info(self) -> BZPlotInfo:
        faces = [[[float(k) for k in j] for j in i]
                 for i in self.rlat.get_wigner_seitz_cell()]
        labels, concat, band_paths, init_point = {}, False, [], None

        for kpoint in self.kpoints:
            if kpoint.label:
                c_coords = list(kpoint.cart_coords)
                f_coords = list(kpoint.frac_coords)
                label = greek_to_unicode(kpoint.label)
                labels[label] = {"cart": c_coords, "frac": f_coords}
                if concat is False and init_point:
                    band_paths.append([init_point, c_coords])
                init_point = c_coords
                concat = True
            else:
                concat = False

        return BZPlotInfo(faces, labels, band_paths, self.rlat.matrix.tolist())

    def _remove_spin_key(self, plot_data):
        """
        Pymatgen at 2020.11.11
         energy: A dict storing bands for spin up and spin down data
            {Spin:[np.array(nb_bands,kpoints),...]} as a list of discontinuous
            kpath of energies. The energy of multiple continuous branches are
            stored together.
        -> [branch][spin][band][k-point]
        """
        num_spin = len(plot_data["energy"])
        num_branch = len(plot_data["energy"]["1"])

        result = [[[] for _ in range(num_spin)] for __ in range(num_branch)]
        for spin_idx, (_, branch_energies) in enumerate(
                sorted(plot_data["energy"].items(),
                       key=lambda item: item[0],
                       reverse=True)):
            for branch_idx, branch_energy in enumerate(branch_energies):
                if self.energy_window:
                    removed_idxs = []
                    for i in range(len(branch_energy)):
                        _max = np.max(branch_energy[i, :])
                        _min = np.min(branch_energy[i, :])
                        if not self.in_energy(_max, _min):
                            removed_idxs.append(i)
                    x = np.delete(branch_energy, removed_idxs, axis=0).tolist()
                else:
                    x = branch_energy.tolist()

                result[branch_idx][spin_idx] = deepcopy(x)

        return result

    def in_energy(self, _max, _min):
        return _max >= self.energy_window[0] and _min <= self.energy_window[1]

    @staticmethod
    def _band_edge(bs, plot_data):
        if bs.is_metal():
            return None
        else:
            return BandEdgeForPlot(
                vbm=plot_data["vbm"][0][1],
                cbm=plot_data["cbm"][0][1],
                vbm_distances=[i[0] for i in plot_data["vbm"]],
                cbm_distances=[i[0] for i in plot_data["cbm"]])

    @property
    def _title(self):
        return latexify(self._composition.reduced_formula)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import os
import re
from collections import OrderedDict
from copy import deepcopy
from pathlib import Path
from typing import List, Dict, Any

from monty.io import zopen
from pymatgen.electronic_structure.core import Magmom
from pymatgen.io.vasp import Incar
from tabulate import tabulate

from vise.input_set.datasets.dataset_util import incar_categories
from vise.util.logger import get_logger

MODULE_DIR = Path(os.path.dirname(os.path.abspath(__file__)))


logger = get_logger(__name__)


class ViseIncar(Incar):
    """Incar class modified for pretty writing of INCAR file. """
    @classmethod
    def from_file(cls, filename: str) -> "ViseIncar":
        """
        Since from_file and from_string methods in Incar class use 'Incar' class
        constructor, we need to override them.
        """
        with zopen(filename, "rt") as f:
            return cls.from_string(f.read())

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "ViseIncar":
        kwargs = deepcopy(d)
        if kwargs.get("MAGMOM") and isinstance(kwargs["MAGMOM"][0], dict):
            kwargs["MAGMOM"] = [Magmom.from_dict(m) for m in kwargs["MAGMOM"]]

        return cls({k: v for k, v in d.items()
                    if k not in ("@module", "@class")})

    @classmethod
    def from_string(cls, string: str) -> "ViseIncar":
        """Reads an Incar object from a string.

        This method is different from that of Incar superclass as it supports
        ";" semantic, which splits the incar flags.
        """
        params = {}
        for line in string.splitlines():
            # Support ";" semantic for split of flags in INCAR file
            for sub_line in line.split(";"):
                m = re.match(r"\s*(\w+)\s*=\s*(.*)\s*", sub_line)
                if m:
                    key = m.group(1).strip()
                    val = m.group(2).strip()
                    params[key] = ViseIncar.proc_val(key, val)
        return cls(params)

    def __add__(self, other: Incar) -> "ViseIncar":
        """Add all the values of another INCAR object to this object. """
        original_settings = {k: v for k, v in self.items()}
        for key, value in other.items():
            if key in self and value != self[key]:
                raise ValueError("INCARs have conflicting values!")
            else:
                original_settings[key] = value
        return ViseIncar(original_settings)

    def get_string(self, **kwargs) -> str:
        lines = OrderedDict()
        input_incar_flags = list(self.keys())

        for category, flags_by_category in incar_categories.items():
            settings_by_category = []

            for flag in flags_by_category:
                if flag in input_incar_flags:
                    settings_by_category.append(
                        [flag, self.setting_to_str(flag)])
                    input_incar_flags.remove(flag)

            if settings_by_category:
                lines[f"# {category}"] = tabulated_string(settings_by_category)

        if input_incar_flags:
            logger.error(f"{input_incar_flags} may be invalid in INCAR.,"
                         f"We are sorry, but add it by hands")

        return "\n\n".join(["\n".join([k, v]) for k, v in lines.items()])

    def get_str(self, **kwargs) -> str:
        return self.get_string(**kwargs)

    def setting_to_str(self, incar_flag: str) -> str:
        if isinstance(self[incar_flag], list):
            return " ".join([str(i) for i in self[incar_flag]])
        else:
            return str(self[incar_flag])

    @property
    def is_ncl_calc(self) -> bool:
        return any([self.get("LNONCOLLINEAR", False),
                    self.get("LSORBIT", False)])


def tabulated_string(target_list: List[list]) -> str:
    """Tabulate pairs of keys and values with sandwiching "=".

    Examples: target_list = [["NSW", 1], ["ISMEAR", -5]]
     ->
NSW    =  1
ISMER  = -5"""
    # if not disable_numparse, e.g., LOPTICS = True becomes 1.
    equal_added = [[key, "=", value] for (key, value) in target_list]
    return str(tabulate(equal_added, tablefmt="plain", disable_numparse=True))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import argparse
import sys
import warnings
from pathlib import Path

from pymatgen.io.vasp.inputs import UnknownPotcarWarning
from vise import __version__
from vise.analyzer.atom_grouping_type import AtomGroupingType
from vise.analyzer.plot_diele_func_data import DieleFuncPlotType
from vise.cli.main_functions import get_poscar_from_mp, VaspSet, plot_band, \
    plot_dos, band_edge_properties, plot_diele_func, \
    calc_effective_mass, structure_info
from vise.defaults import defaults
from vise.input_set.task import Task
from vise.input_set.xc import Xc
from vise.util.logger import get_logger
from vise.util.str_related_tools import str2bool

logger = get_logger(__name__)


warnings.simplefilter('ignore', UnknownPotcarWarning)


description = """Vise is a package that helps researchers to
to generate VASP input files for several tasks with suitable defaults, 
and allows for tuning some parameters depending on their particular purposes."""

epilog = f"Author: Yu Kumagai Version: {__version__}"


def parse_args(args):
    parser = argparse.ArgumentParser(
        description=description, epilog=epilog,
        formatter_class=argparse.RawTextHelpFormatter)

    subparsers = parser.add_subparsers()

    # -- parent parser: vasprun
    vasprun_parser = argparse.ArgumentParser(description="", add_help=False)
    vasprun_parser.add_argument(
        "-v", "--vasprun", type=Path, default=defaults.vasprun,
        help="vasprun.xml file name.")
    # -- parent parser: outcar
    outcar_parser = argparse.ArgumentParser(description="", add_help=False)
    outcar_parser.add_argument(
        "-o", "--outcar", type=Path, default=defaults.outcar,
        help="OUTCAR file name.")

    # -- structure_info ---------------------------------------------------
    parser_structure_info = subparsers.add_parser(
        name="structure_info",
        description="Shows structure info or conventional cell.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['si'])

    parser_structure_info.add_argument(
        "-p", "--poscar", type=str, default="POSCAR",
        help="POSCAR-type file name to be read.", metavar="FILE")
    parser_structure_info.add_argument(
        "-s", "--symprec", type=float,
        default=defaults.symmetry_length_tolerance,
        help="Length tolerance in Å for symmetry analysis.")
    parser_structure_info.add_argument(
        "-a", "--angle_tolerance", type=float,
        default=defaults.symmetry_angle_tolerance,
        help="Angle tolerance in degree for symmetry analysis.")
    parser_structure_info.add_argument(
        "-c", "--show_conventional", action="store_true",
        help="Describe the conventional cell.")
    parser_structure_info.add_argument(
        "-primitive", "--show_primitive", action="store_true",
        help="Describe the primitive cell.")

    parser_structure_info.set_defaults(func=structure_info)

    # -- get_poscar -----------------------------------------------------------
    parser_get_poscar = subparsers.add_parser(
        name="get_poscar",
        description="Retrieves a POSCAR file from Materials Project Database.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['gp'])

    parser_get_poscar.add_argument(
        "-m", "--mpid", type=str,
        help="MP entry id with prefix, e.g., mp-1234.")
    parser_get_poscar.add_argument(
        "-f", "--formula", type=str,
        help="Formula, e.g., SrTiO3. POSCAR with the lowest energy is "
             "retrieved. ")

    parser_get_poscar.set_defaults(func=get_poscar_from_mp)

    # -- vasp_set ---------------------------------------------------------
    parser_vasp_set = subparsers.add_parser(
        name="vasp_set",
        description="Constructs vasp input set",
        parents=[vasprun_parser, outcar_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['vs'])

    parser_vasp_set.add_argument(
        "-d", "--dirs", nargs="+", type=Path, default=[Path.cwd()],
        help="Directory paths to be parsed.")
    parser_vasp_set.add_argument(
        "-p", "--poscar", default=None, type=Path,
        help="POSCAR-type input structure file name.")
    parser_vasp_set.add_argument(
        "-t", dest="task", default=defaults.task, type=Task,
        choices=Task.name_list(),
        help=f"Task name.")
    parser_vasp_set.add_argument(
        "-x", dest="xc", default=defaults.xc, type=Xc, choices=Xc.name_list(),
        help=f"Exchange-correlation (XC) interaction.")
    parser_vasp_set.add_argument(
        # Default must be None to use the insulator_kpoint_density.
        # defaults.kpoints_density is used at structure_kpoints_generator.py
        "-k", "--kpt_density", type=float,
        help="K-point density in Angstrom along each direction. "
             f"Defaults of metal and insulators are {defaults.kpoint_density}, "
             f"and {defaults.insulator_kpoint_density}.")
    parser_vasp_set.add_argument(
        "--potcar", dest="overridden_potcar",
        default=defaults.overridden_potcar, type=str, nargs="+",
        help="User specifying POTCAR set. E.g., Mg_pv O_h")
    parser_vasp_set.add_argument(
        "-uis", "--user_incar_settings", type=str, nargs="+",
        help="""Overwritten incar settings described with pairs of INCAR tag 
        names and values. These can also be set by vise.yaml, but they are 
        overwritten by this command line options.""")
    parser_vasp_set.add_argument(
        "-pd", "--prev_dir", type=Path,
        help="Parse the previous calculations for better input settings.")
    parser_vasp_set.add_argument(
        "--options", type=str, nargs="+",
        help="Manual options used at CategorizedInputOptions class.")
    parser_vasp_set.add_argument(
        "--uniform_kpt_mode", action="store_true",
        help="""Kpoints with uniform k-point sampling. The k-point sampling mesh 
        and centering are determined based on the given lattice. Note that only 
        when the angles are 90 degrees, the centering is shifted along the 
        perpendicular direction. This mode is useful when calculating the 
        supercells. """)
    parser_vasp_set.add_argument(
        "--file_transfer", dest="file_transfer_type", type=str, nargs="+",
        help="File transfer which is used along with the prev_dir argument and "
             "can be written with pairs of file names and transfer types of "
             "m (move), c (copy) or l (link). An example is: "
             "--file_transfer_type POSCAR c WAVECAR l ")

    parser_vasp_set.set_defaults(func=VaspSet)

    # -- plot_band -------------------------------------------------------------
    parser_plot_band = subparsers.add_parser(
        name="plot_band",
        description="Plots a band structure",
        parents=[vasprun_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pb'])

    parser_plot_band.add_argument(
        "-k", "--kpoints", dest="kpoints_filename", default="KPOINTS", type=str,
        help="KPOINTS file name.")
    parser_plot_band.add_argument(
        "-y", "--y_range", nargs="+", default=[-10.0, 10.0], type=float,
        help="Energy range, requiring two values.")
    # parser_plot_band.add_argument(
    #     "-w", "--wavecar", dest="wavecar_filename", default=None, type=str,
    #     help="WAVECAR file name.")
    parser_plot_band.add_argument(
        "-p", "--poscar", type=str, default="POSCAR",
        help="POSCAR-type file name to be read, which is used for evaluating "
             "the irreps.", metavar="FILE")
    parser_plot_band.add_argument(
        "-f", "--filename", type=str, default="band.pdf", help="Pdf file name.")

    parser_plot_band.set_defaults(func=plot_band)

    # -- plot_dos -----------------------------------------------------------
    parser_plot_dos = subparsers.add_parser(
        name="plot_dos",
        description="Plots density of states",
        parents=[vasprun_parser, outcar_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pd'])

    parser_plot_dos.add_argument(
        "-t", "--type",
        type=AtomGroupingType,
        default=AtomGroupingType.non_equiv_sites,
        choices=AtomGroupingType.name_list(),
        help=f"How to group atoms for pdos. When {AtomGroupingType.atoms} is "
             f"chosen, the atom indices need to be set via --target.")
    parser_plot_dos.add_argument(
        "-l", "--legend", type=str2bool, default=True,
        help="Whether to show the figure legend.")
    parser_plot_dos.add_argument(
        "-c", "--crop_first_value", type=str2bool, default=True,
        help="Whether to crop the first value in DOS.")
    parser_plot_dos.add_argument(
        "-x", "--x_range", nargs="+", type=float,
        help="Set minimum and maximum energies.")
    parser_plot_dos.add_argument(
        "-y", "--y_max_ranges", nargs="+", type=float,
        help="Set max energies for each plot. AssertionError is raised when"
             "the number of ranges does not fit to the number of figures.")
    parser_plot_dos.add_argument(
        "--target", type=str, nargs="+",
        help="""Show specific PDOS. The input depends on AtomGroupingType.\n 
        AtomGroupingType.atoms:   1,2 3 \n
        AtomGroupingType.elements: Mg \n
        Note that the indices starts from 0.""")
    parser_plot_dos.add_argument(
        "-f", "--filename", type=str, default="dos.pdf", help="Pdf file name.")
    parser_plot_dos.add_argument(
        "-b", "--base_energy", type=float,
        help="Set when showing the figure in the absolute energies scale.")
    parser_plot_dos.add_argument(
        "--title", type=str,
        help="Set the title of figure.")
    parser_plot_dos.set_defaults(func=plot_dos)

    # -- plot_diele_func -------------------------------------------------------
    parser_plot_absorption = subparsers.add_parser(
        name="plot_diele_func",
        description="Plots dielectric function and its related quantities such "
                    "as optical absorption coefficient.",
        parents=[vasprun_parser, outcar_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pdf'])

    parser_plot_absorption.add_argument(
        "-f", "--filename", type=str, default=None,
        help="Output PDF file name.")
    parser_plot_absorption.add_argument(
        "--input_csv_name", type=str, default=None,
        help="Input CSV file name for dielectric function.")
    parser_plot_absorption.add_argument(
        "-d", "--directions", nargs="+", type=str,
        default=["ave"],
        choices=["xx", "yy", "zz", "xy", "yz", "xz", "ave"],
        help="Directions for the plot: For example,set 'xx' for x-direction.")
    parser_plot_absorption.add_argument(
        "-y", "--y_range", nargs="+", default=None,
        type=float,
        help="Requiring two values. For absorption coefficient, exponential "
             "parts of base-10 for energy range in cm-1.")
    parser_plot_absorption.add_argument(
        "-ckk", "--calc_kk", action="store_true",
        help="Set when real part of dielectric function is explicitly "
             "calculated using the Kramers-Kronig transformation.")
    parser_plot_absorption.add_argument(
        "-i", "--ita", type=float, default=0.01,
        help="Complex shift η in the Kramers-Kronig transformation.")
    parser_plot_absorption.add_argument(
        "--plot_type", type=DieleFuncPlotType,
        default=DieleFuncPlotType.absorption_coeff,
        choices=DieleFuncPlotType.name_list(),
        help="Choose which type of data is drawn.")
    parser_plot_absorption.add_argument(
        "--to_csv", action="store_true",
        help="Whether to export the dielectric function to CSV file.")
    parser_plot_absorption.add_argument(
        "--title", type=str,
        help="Set the title of figure.")
    parser_plot_absorption.set_defaults(func=plot_diele_func)

    # -- effective_mass -------------------------------------------------------
    parser_effective_mass = subparsers.add_parser(
        name="effective_mass",
        description="Calculating effective masses via BoltzTrap2."
                    "When using this, one needs to cite its paper(s).",
        parents=[vasprun_parser, outcar_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['em'])

    parser_effective_mass.add_argument(
        "-t", "--temperature", type=float, default=300,
        help="Temperature in K.")
    parser_effective_mass.add_argument(
        "-c", "--concentrations", type=lambda x: 10 ** float(x), nargs="+",
        help="Exponential parts of base-10 for concentrations in cm-3")

    parser_effective_mass.set_defaults(func=calc_effective_mass)

    # -- band_edge -------------------------------------------------------------
    parser_band_edge = subparsers.add_parser(
        name="band_edge",
        description="Calculates the band edge positions",
        parents=[vasprun_parser, outcar_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['be'])

    parser_band_edge.set_defaults(func=band_edge_properties)

    try:
        import argcomplete
        argcomplete.autocomplete(parser)
        # https://github.com/kislyuk/argcomplete#activating-global-completion
        # This supports bash/zsh autocompletion. To enable this, pip install
        # argcomplete, activate global completion, or add
        #      eval "$(register-python-argcomplete vise)"
        # into your .bash_profile or .bashrc
    except ImportError:
        pass

    return parser.parse_args(args)


def main():
    args = parse_args(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import json
from math import ceil
from pathlib import Path
from typing import Tuple, List, Dict, Any

from monty.serialization import loadfn
from pymatgen.core import Composition, Element
from pymatgen.io.vasp import Potcar
from vise.defaults import defaults

from pymatgen.io.vasp import inputs

p = Path(inputs.__file__).parent
with open(p / "incar_parameters.json", encoding="utf-8") as json_file:
    incar_params = json.loads(json_file.read())

unoccupied_bands = loadfn(Path(__file__).parent / "unoccupied_bands.yaml")

# This incar_flags should be OrderedDict, but from python 3.6, dict uses
# order-preserving semantics. Besides, it does not affect vasp result.
incar_categories: Dict[str, Any] = \
    dict(loadfn(Path(__file__).parent / "incar_flags.yaml"))
tag_set = set(tuple(tags) for tags in incar_categories)
incar_categories["others"] = list(set(incar_params.keys()) - tag_set)
all_incar_flags: List[str] = sum(incar_categories.values(), [])


def has_f_elements(symbol_list: List[str]):
    return any([Element(el).Z > 56 for el in symbol_list])


class LDAU:
    def __init__(self,
                 symbol_list: List[str]):

        self.symbol_list = symbol_list
        ldau_set = loadfn(defaults.u_parameter_set_yaml_file)
        self.ldauu = [ldau_set["LDAUU"].get(el, 0) for el in symbol_list]
        self.ldaul = [ldau_set["LDAUL"].get(el, -1) for el in symbol_list]

    @property
    def lmaxmix(self):
        return 6 if has_f_elements(self.symbol_list) else 4

    @property
    def is_ldau_needed(self) -> bool:
        return set(self.ldauu) != {0}


def num_bands(composition: Composition, potcar: Potcar,
              spin_orbit: bool = False) -> int:
    """Required for optical absorption, band structure, and DOS"""
    results = 0
    for element, potcar_single in zip(composition, potcar):
        num_atoms_per_element = composition[element]
        occupied_bands = potcar_single.nelectrons / 2
        num_bands_per_atom = occupied_bands + unoccupied_bands[str(element)]
        results += num_atoms_per_element * num_bands_per_atom

    if spin_orbit:
        results *= 2

    return ceil(results)


def calc_kpar(num_kpoints: int, num_cores: int,
              unused_core_ratio_threshold: float) -> int:
    divisors = [i for i in range(num_cores, 0, -1) if num_cores % i == 0]
    for d in divisors:
        kpt_by_core = round(num_kpoints / d, 5)
        frac_kpt_by_core = kpt_by_core % 1
        frac_kpt_by_core = 0.0 if not frac_kpt_by_core else 1 - frac_kpt_by_core
        unused_core_ratio = frac_kpt_by_core / ceil(kpt_by_core)
        if unused_core_ratio < unused_core_ratio_threshold:
            return d

    raise ValueError(f"The threshold for unused core ratio "
                     f"{unused_core_ratio_threshold} is not adequate.")





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from unittest.mock import MagicMock

import pytest

from vise.analyzer.plot_band import (
    BandMplPlotter, BandEnergyInfo, BandEdgeForPlot, XTicks, BandMplSettings,
    BandPlotInfo, ViseBandInfoError)
from vise.tests.helpers.assertion import assert_msonable
from vise.util.matplotlib import float_to_int_formatter

"""
TODO:
+ To csv
"""

distances = [[0, 1, 2, 3, 4, 5, 6], [6, 7, 8, 9]]
x_ticks = XTicks(["A", "B", "C$\\mid$D", "E"], [0.0, 5.0, 6.0, 9.0])
y_range = [-10, 10]
title = "Title"
base_energy = -1.0
# [by branch][by spin][by band idx][by k-path idx]
# two branches, no spin, single band, and 7 and 4 k-points in two branches.
# irrep exists for first k-point in the first branch.


@pytest.fixture
def band_energies():
    return [[[[-3.0, -2, -1, -1, -1, -2, -3], [7.0, 6, 5, 4, 3, 2, 3]]],
            [[[-4.0, -5, -6, -7], [5, 2, 7, 8]]]]


shifted_band_energies = \
    [[[[-2.0, -1.0, 0.0, 0.0, 0.0, -1.0, -2.0],
       [8.0, 7.0, 6.0, 5.0, 4.0, 3.0, 4.0]]],
     [[[-3.0, -4.0, -5.0, -6.0], [6.0, 3.0, 8.0, 9.0]]]]

fermi_level = 1.5


colors = ['#E15759', '#4E79A7', '#F28E2B', '#76B7B2']


@pytest.fixture
def band_edge():
    return BandEdgeForPlot(vbm=-1, cbm=2,
                           vbm_distances=[2, 3, 4], cbm_distances=[5, 7])


@pytest.fixture
def band_energy_info_wo_fermi(band_energies, band_edge):
    return BandEnergyInfo(band_energies=band_energies, band_edge=band_edge)


@pytest.fixture
def band_energy_info(band_edge, band_energies, irreps):
    return BandEnergyInfo(band_energies=band_energies,
                          band_edge=band_edge, fermi_level=fermi_level,
                          irreps=irreps)


@pytest.fixture
def band_infos(band_energy_info):
    return {"subtitle": band_energy_info}


@pytest.fixture
def band_plot_info(band_infos):
    return BandPlotInfo(band_infos, distances, x_ticks, title)


def test_irrep_msonable(irreps):
    assert_msonable(irreps)


def test_irrep_info_set(irreps):
    assert list(irreps()["Γ"].irrep_info_set) == [("Γ1+", 0.1, 1)]


def test_get_distances(irreps):
    actual = irreps.get_distances(XTicks(labels=["Γ"], distances=[0.0]))
    assert actual == [[0.0]]


def test_band_plot_info_msonable(band_edge, band_energy_info, band_plot_info):
    assert_msonable(band_edge)
    assert_msonable(band_energy_info)
    assert_msonable(band_plot_info)


@pytest.fixture
def mock_band_plt_list(mocker, band_plot_info):
    mock_plt = mocker.patch("vise.analyzer.plot_band.plt", unsafe=True)
    mock_axis = MagicMock()
    mock_plt.gca.return_value = mock_axis
    plotter = BandMplPlotter(band_plot_info, y_range)
    plotter.construct_plot()
    return mock_plt, mock_axis


def test_band_info_slide_energies(band_energy_info):
    band_energy_info.slide_energies(base_energy=base_energy)
    expected_band_edge = BandEdgeForPlot(
        vbm=0, cbm=3, vbm_distances=[2, 3, 4], cbm_distances=[5, 7])
    assert band_energy_info.band_energies == shifted_band_energies
    assert band_energy_info.band_edge == expected_band_edge
    assert band_energy_info.fermi_level == fermi_level - base_energy
    assert band_energy_info.irreps.irreps["Γ"].energies[0] == 1.1


def test_raise_error_when_both_band_edge_fermi_level_absent(band_energies):
    # [spin][branch idx][band idx][k-path idx]
    with pytest.raises(ViseBandInfoError):
        BandEnergyInfo(band_energies=band_energies)


def test_slide_energies_when_fermi_is_none(band_energy_info_wo_fermi):
    band_energy_info_wo_fermi.slide_energies(base_energy=-1.0)
    assert band_energy_info_wo_fermi.band_edge.vbm == 0.0
    assert band_energy_info_wo_fermi.fermi_level is None


def test_band_plot_info_band_energy_region():
    band_info = BandEnergyInfo(
        band_energies=[[[[-1.01, -1.008, -1.003, -1.0]]],
                       [[[1.01, 1.0]]]],
        band_edge=BandEdgeForPlot(vbm=-1.0, cbm=1.0,
                                  vbm_distances=[1], cbm_distances=[1]),
        fermi_level=0.0)
    assert band_info.band_energy_region() == [[-1.01, -1.0], [1.0, 1.01]]
    assert band_info.band_energy_region(decision_width=0.0031) \
           == [[-1.01, -1.008], [-1.003, -1.0], [1.0, 1.0], [1.01, 1.01]]
    assert band_info.band_energy_region(bottom=0.0) == [[1.0, 1.01]]
    assert band_info.band_energy_region(top=0.0) == [[-1.01, -1.0]]
    assert band_info.band_energy_region(offset=0.1) == [[-1.11, -1.1],
                                                        [0.9, 0.91]]


def test_band_mpl_defaults():
    band_defaults = BandMplSettings()

    assert band_defaults.colors == colors
    assert next(band_defaults.linewidth) == 1.0
    assert band_defaults.circle_size == 70
    assert band_defaults.circle_colors == ["pink", "green"]
    assert band_defaults.title_font_size == 15
    assert band_defaults.label_font_size == 15

    band_defaults = BandMplSettings(colors=["black"],
                                    linewidth=[2.0],
                                    circle_size=200,
                                    circle_colors=["blue"],
                                    title_font_size=30,
                                    label_font_size=40)

    assert band_defaults.colors == ["black"]
    assert next(band_defaults.linewidth) == 2.0
    assert band_defaults.circle_size == 200
    assert band_defaults.circle_colors == ["blue"]
    assert band_defaults.title_font_size == 30
    assert band_defaults.label_font_size == 40


def test_add_band_structures(mock_band_plt_list):
    mock_plt, _ = mock_band_plt_list
    linewidth = BandMplSettings().linewidth
    # 1st branch, 1st band
    args = {"color": colors[0],
            "linewidth": next(linewidth),
            'label': 'subtitle'}
    plotted_band_energies = \
    [[[[-2.0, -1, 0, 0, 0, -1, -2],
       [8.0, 7, 6, 5, 4, 3, 4]]],
     [[[-3.0, -4, -5, -6], [6, 3, 8, 9]]]]
    print(distances[0], plotted_band_energies[0][0][0], args)
    mock_plt.plot.assert_any_call(distances[0], plotted_band_energies[0][0][0],
                                  **args)
    # 2nd branch, 1st band
    args.pop("label", None)
    mock_plt.plot.assert_any_call(distances[1], plotted_band_energies[1][0][0],
                                  **args)
    # 1st branch, 2nd band
    mock_plt.plot.assert_any_call(distances[0], plotted_band_energies[0][0][1],
                                  **args)
    # 2nd branch, 2nd band
    mock_plt.plot.assert_any_call(distances[1], plotted_band_energies[1][0][1],
                                  **args)


def test_band_plot_info_add(band_plot_info, band_energy_info: BandEnergyInfo):
    band_plot_info_2 = BandPlotInfo({"a": band_energy_info}, distances, x_ticks)
    added = band_plot_info + band_plot_info_2
    assert added.band_energy_infos["a"].as_dict() == band_energy_info.as_dict()
    assert added.distances_by_branch == distances
    assert added.x_ticks == x_ticks


def test_add_band_edge_circles(mock_band_plt_list, band_infos):
    mock_plt, _ = mock_band_plt_list
    edge = band_infos["subtitle"].band_edge
    defaults = BandMplSettings()

    mock_plt.scatter.assert_any_call(edge.vbm_distances[0], 0,
                                     **defaults.circle(0))
    mock_plt.scatter.assert_any_call(edge.cbm_distances[0], 3.0,
                                     **defaults.circle(0))
    mock_plt.scatter.assert_any_call(edge.cbm_distances[1], 3.0,
                                     **defaults.circle(0))
    mock_plt.axhline.assert_any_call(y=0.0, **defaults.hline)
    mock_plt.axhline.assert_any_call(y=3.0, **defaults.hline)


def test_figure_legends(mock_band_plt_list):
    mock_plt, _ = mock_band_plt_list
#    mock_plt.legend.assert_called_once_with(loc="lower right")
    mock_plt.legend.assert_not_called()


def test_set_x_range(mock_band_plt_list):
    mock_plt, _ = mock_band_plt_list
    mock_plt.xlim.assert_called_with(distances[0][0], distances[-1][-1])


def test_set_y_range(mock_band_plt_list):
    mock_plt, _ = mock_band_plt_list
    mock_plt.ylim.assert_called_with(y_range[0], y_range[1])


def test_set_labels(mock_band_plt_list):
    mock_plt, _ = mock_band_plt_list
    defaults = BandMplSettings()
    mock_plt.xlabel.assert_called_with("Wave vector",
                                       size=defaults.label_font_size)
    mock_plt.ylabel.assert_called_with("Energy (eV)",
                                       size=defaults.label_font_size)


def test_set_x_tics(mock_band_plt_list):
    mock_plt, mock_axis = mock_band_plt_list

    mock_axis.set_xticks.assert_called_once_with(x_ticks.distances)
    mock_axis.set_xticklabels.assert_called_once_with(x_ticks.labels)
    mock_plt.axvline.assert_any_call(x=x_ticks.distances[2], linestyle="-")
    mock_plt.axvline.assert_any_call(x=x_ticks.distances[1], linestyle="--")


def test_set_title(mock_band_plt_list):
    mock_plt, _ = mock_band_plt_list
    defaults = BandMplSettings()
    mock_plt.title.assert_called_once_with("Title",
                                           size=defaults.title_font_size)


def test_set_float_to_int_formatter(mock_band_plt_list):
    _, mock_axis = mock_band_plt_list
    print(mock_axis)
    mock_axis.yaxis.set_major_formatter.assert_called_once_with(
        float_to_int_formatter)


def test_plot_tight_layout(mock_band_plt_list):
    mock_plt, _ = mock_band_plt_list
    mock_plt.tight_layout.assert_called_once_with()


def test_draw_bands(band_plot_info):
    band_plotter = BandMplPlotter(band_plot_info, y_range)
    band_plotter.construct_plot()
    band_plotter.plt.show()


@pytest.mark.parametrize("ref_energy,subtracted_energy",
                         [(1.0, 1.0), (None, -1)])
def test_reference_energy(ref_energy, subtracted_energy,
                          mocker, band_plot_info, band_edge):
    mock_plt = mocker.patch("vise.analyzer.plot_band.plt", unsafe=True)
    mock_axis = MagicMock()
    mock_plt.gca.return_value = mock_axis

    plotter = BandMplPlotter(band_plot_info, y_range, ref_energy)
    plotter.construct_plot()

    band_defaults = BandMplSettings()
    mock_plt.axhline.assert_called_with(y=band_edge.cbm - subtracted_energy,
                                        **band_defaults.hline)


@pytest.fixture
def two_band_infos():
    first_branch_energies = [[[-3.0, -2, -1, -1, -1, -2, -3],
                               [7.0, 6, 5, 4, 3, 2, 3]],
                              [[-2.0, -1, 0, 0, 0, -1, -2],
                               [6.0, 5, 4, 3, 2, 1, 2]]]
    second_branch_energies = [[[-4, -5, -6, -7],
                               [5, 2, 7, 8]],
                              [[-3, -4, -5, -6],
                               [4, 1, 6, 7]]]
    first_band_energies = [first_branch_energies, second_branch_energies]

    band_edge_1 = BandEdgeForPlot(vbm=0, cbm=1,
                                  vbm_distances=[2, 3, 4], cbm_distances=[5, 7])
    band_info_1 = BandEnergyInfo(band_energies=first_band_energies,
                                 band_edge=band_edge_1,
                                 fermi_level=None)

    first_branch_energies = [[[-5, -5, -5, -5, -4, -5, -5],
                              [5, 5, 5, 5, 4, 5, 5]],
                             [[-6, -6, -6, -6, -5, -6, -6],
                              [6, 6, 6, 6, 5, 6, 6]]]
    second_branch_energies = [[[-5, -5, -5, -5],
                               [5, 5, 5, 5]],
                              [[-6, -6, -6, -6],
                               [6, 6, 6, 6]]]
    second_band_energies = [first_branch_energies, second_branch_energies]

    band_edge_2 = BandEdgeForPlot(vbm=-4, cbm=4,
                                  vbm_distances=[4], cbm_distances=[4])
    band_info_2 = BandEnergyInfo(band_energies=second_band_energies,
                                 band_edge=band_edge_2,
                                 fermi_level=None)

    return {"1": band_info_1, "2": band_info_2}


@pytest.fixture
def mock_two_bands_plt_axis(two_band_infos, mocker):
    mock_plt = mocker.patch("vise.analyzer.plot_band.plt", unsafe=True)
    mock_axis = MagicMock()
    mock_plt.gca.return_value = mock_axis
    band_plot_info = BandPlotInfo(two_band_infos, distances, x_ticks, title)
    plotter = BandMplPlotter(band_plot_info, y_range)
    plotter.construct_plot()
    return mock_plt, mock_axis


def test_check_color_generator_with_last_call(mock_two_bands_plt_axis):
    mock_plt, _ = mock_two_bands_plt_axis
    mock_plt.plot.assert_called_with(
        distances[1],
        [6, 6, 6, 6],
        color=colors[1],
        linewidth=1.0,
        linestyle=":")


def test_draw_two_bands(two_band_infos):
    band_plot_info = BandPlotInfo(two_band_infos, distances, x_ticks, title)
    band_plotter = BandMplPlotter(band_plot_info, y_range)
    band_plotter.construct_plot()
    band_plotter.plt.show()




# -*- coding: utf-8 -*-
#  Copyright (c) 2021. Distributed under the terms of the MIT License.
from pymatgen.io.vasp import PotcarSingle


def valence_orbitals_from_potcar(potcar_single: PotcarSingle):

    result = []
    for config in potcar_single.electron_configuration:
        principal_num, orb_type, num_ele = config
        result.append(f"({principal_num}{orb_type}){num_ele}")

    return " ".join(result)





# -*- coding: utf-8 -*-
#  Copyright (c) 2022. Distributed under the terms of the MIT License.
from copy import copy
from dataclasses import dataclass
from typing import Optional, Dict

import yaml
from monty.json import MSONable
from vise.input_set.datasets.potcar_set import PotcarSet
from vise.input_set.kpoints_mode import KpointsMode
from vise.input_set.task import Task
from vise.input_set.xc import Xc
from vise.util.mix_in import ToYamlFileMixIn


@dataclass
class ViseLog(MSONable, ToYamlFileMixIn):
    version: str
    task: Task
    xc: Xc
    input_options: dict
    user_incar_settings: Optional[dict] = None
    ldauu: Optional[Dict[str, float]] = None
    ldaul: Optional[Dict[str, float]] = None

    def to_yaml(self):
        return yaml.safe_dump(self.as_dict(), sort_keys=False)

    def as_dict(self) -> dict:
        d = {"version": self.version, "task": str(self.task),
             "xc": str(self.xc), "input_options": self.input_options}
        if self.user_incar_settings is not None:
            d["user_incar_settings"] = self.user_incar_settings
        if self.ldauu is not None:
            d["ldauu"] = self.ldauu
        if self.ldaul is not None:
            d["ldaul"] = self.ldaul
        if "kpt_mode" in self.input_options:
            self.input_options["kpt_mode"] = str(self.input_options["kpt_mode"])
        if "potcar_set" in self.input_options:
            self.input_options["potcar_set"] = \
                str(self.input_options["potcar_set"])
        return d

    @classmethod
    def from_dict(cls, d):
        dd = copy(d)
        dd["task"] = Task.from_string(d["task"])
        dd["xc"] = Xc.from_string(d["xc"])
        if "kpt_mode" in d["input_options"]:
            dd["input_options"]["kpt_mode"] = \
                KpointsMode.from_string(dd["input_options"]["kpt_mode"])
        if "potcar_set" in d["input_options"]:
            dd["input_options"]["potcar_set"] = \
                PotcarSet.from_string(dd["input_options"]["potcar_set"])

        return super().from_dict(dd)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from argparse import Namespace
from pathlib import Path

from pymatgen.io.vasp import Chgcar
from vise.analyzer.vasp.handle_volumetric_data import \
    light_weight_vol_text, make_spin_charges
from vise.analyzer.vesta.vesta_file import add_density, VestaFile, calc_isurfs
from vise.atom_energies.make_atom_vasp_set import make_atom_poscar_dirs
from vise.util.logger import get_logger
from vise.util.phonopy.phonopy_input import make_phonopy_input
from vise.util.structure_handler import sanitize_matrix

logger = get_logger(__name__)


def make_phonon_poscars(args: Namespace) -> None:
    phonopy_input = make_phonopy_input(
        unitcell=args.unitcell,
        supercell_matrix=sanitize_matrix(args.supercell_matrix))
    phonopy_input.supercell.to(filename="POSCAR")
    phonopy_input.to_json_file()


def make_phonon_figs(args: Namespace) -> None:
    args.phonopy_input.set_force_constants_from_vasprun(args.vasprun_name)
    phonopy = args.phonopy_input.to_phonopy
    plt = phonopy.auto_band_structure(plot=True)
    plt.savefig("phonon_band.pdf")
    plt.show()


def make_atom_poscars(args: Namespace) -> None:
    make_atom_poscar_dirs(args.dirname, args.elements)


def make_spin_decomposed_volumetric_files(args):
    chgcar = Chgcar.from_file(args.chgcar)
    for c, spin in zip(make_spin_charges(chgcar), ["up", "down"]):
        c.write_file(f"{args.chgcar}_{spin}")


def make_light_weight_vol_data(args):
    chgcar = Chgcar.from_file(args.volumetric_file)
    output_vol_filename = (args.output_lw_volumetric_filename
                           or Path(f"{args.volumetric_file}_lw"))
    lw_vol_text = light_weight_vol_text(chgcar, args.border_fractions)
    Path(output_vol_filename).write_text(lw_vol_text)

    if args.output_vesta_filename:
        is_chg = "CHG" in args.volumetric_file
        volume = chgcar.structure.volume
        isurfs = calc_isurfs(args.border_fractions, is_chg, volume)

        if args.original_vesta_file:
            vesta_text = args.original_vesta_file.read_text()
        else:
            vesta_text = VestaFile(chgcar.structure).__repr__()
        to_vesta_text = add_density(vesta_text, isurfs, str(output_vol_filename))
        Path(args.output_vesta_filename).write_text(to_vesta_text)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from typing import Optional

from pymatgen.io.vasp.sets import Potcar

from vise.error import ViseError
from vise.input_set.datasets.potcar_set import PotcarSet
from vise.input_set.xc import Xc
from vise.util.logger import get_logger
from vise.defaults import defaults
logger = get_logger(__name__)


def generate_potcar(symbol_list: list,
                    xc: Xc,
                    potcar_set: PotcarSet = defaults.potcar_set,
                    overridden_potcar: Optional[dict] = None):
    potcar_dict = potcar_set.overridden_potcar_dict(overridden_potcar)
    try:
        potcar_symbols = [potcar_dict[el] for el in symbol_list]
    except KeyError as e:
        raise ViseNoPotcarError(e)

    functional = defaults.lda_potcar if xc == Xc.lda else defaults.gga_potcar
    return Potcar(potcar_symbols, functional=functional)


class ViseNoPotcarError(ViseError):
    pass




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import os

import pytest

from vise.defaults import UserSettings


@pytest.fixture
def user_settings(tmpdir):
    second_last_dir = tmpdir
    os.chdir(second_last_dir)
    second_last_dir.join("test.yaml").write("""
        float: 0.01
        dict:
          d1: 5
          d2: 3""")

    terminal_dir = tmpdir.join("terminal_dir")
    terminal_dir.mkdir()
    os.chdir("terminal_dir")
    terminal_dir.join("test.yaml").write("""
        str: s1 s2
        dict:
          d1: 10 
          d3: 4""")

    return UserSettings("test.yaml"), second_last_dir, terminal_dir


def test_get_yaml_filenames(user_settings):
    user_settings, second_last_dir, terminal_dir = user_settings
    actual = user_settings.yaml_files_from_root_dir
    expected = [second_last_dir / "test.yaml", terminal_dir / "test.yaml"]

    assert actual == expected


def test_overridden_user_settings(user_settings):
    user_settings, _, _ = user_settings
    actual = user_settings.user_settings
    expected = {"float": 0.01,
                "dict": {"d1": 10, "d2": 3, "d3": 4},
                "str": "s1 s2"}
    assert actual == expected


def test_parse_hidden_file(tmpdir):
    os.chdir(tmpdir)
    tmpdir.join("test.yaml").write("""
    float: 0.01""")
    user_settings = UserSettings(yaml_filename="test.yaml")
    actual = user_settings.user_settings
    expected = {"float": 0.01}
    assert actual == expected


def test_add_absolute_path(tmpdir):
    os.chdir(tmpdir)
    tmpdir.join("test.yaml").write("""
    path: a/b""")
    user_settings = UserSettings(yaml_filename="test.yaml")
    actual = user_settings.user_settings["path"]
    expected = tmpdir / "a/b"
    assert actual == expected



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from argparse import Namespace
from pathlib import Path

from pymatgen.core import Element
from vise.analyzer.vasp.handle_volumetric_data import default_border_fractions
from vise.cli.main_util import parse_args

parent_dir = Path(__file__).parent


def test_make_atom_poscars_wo_options():
    parsed_args = parse_args(["map"])
    expected = Namespace(
        dirname=Path.cwd(),
        elements=None,
        func=parsed_args.func)
    assert parsed_args == expected


def test_make_atom_poscars_w_options():
    parsed_args = parse_args(["map", "-d", "a", "-e", "H", "He"])
    expected = Namespace(
        dirname=Path("a"),
        elements=[Element.H, Element.He],
        func=parsed_args.func)
    assert parsed_args == expected


def test_make_spin_decomposed_volumetric_files():
    parsed_args = parse_args(["sdvf", "-c", "CHGCAR"])
    expected = Namespace(chgcar="CHGCAR", func=parsed_args.func)
    assert parsed_args == expected


def test_make_light_weight_vol_data_wo_options():
    parsed_args = parse_args(["lwvd", "-v", "CHGCAR"])
    expected = Namespace(volumetric_file="CHGCAR",
                         output_lw_volumetric_filename=None,
                         border_fractions=default_border_fractions,
                         output_vesta_filename=None,
                         original_vesta_file=None,
                         func=parsed_args.func)
    assert parsed_args == expected


def test_make_light_weight_vol_data_w_options():
    parsed_args = parse_args(["lwvd",
                             "-v", "AECCAR",
                              "-o", "AECCAR_lw",
                              "-b", "0.1",
                              "-out_vesta", "out.vesta",
                              "-orig_vesta", "orig.vesta"])
    expected = Namespace(volumetric_file="AECCAR",
                         output_lw_volumetric_filename=Path("AECCAR_lw"),
                         border_fractions=[0.1],
                         output_vesta_filename=Path("out.vesta"),
                         original_vesta_file=Path("orig.vesta"),
                         func=parsed_args.func)
    assert parsed_args == expected



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pymatgen.io.vasp.sets import Kpoints


class ViseKpoints(Kpoints):

    def __str__(self):
        lines = [self.comment, str(self.num_kpts), self.style.name]
        style = self.style.name.lower()[0]
        if style == "l":
            lines.append(self.coord_type)
        for i in range(len(self.kpts)):
            # The following four lines are modified
            if len(self.kpts) == 1:
                lines.append(" ".join([str(x) for x in self.kpts[i]]))
            else:
                lines.append(" ".join([f"{x:20.17f}" for x in self.kpts[i]]))
            if style == "l":
                lines[-1] += " ! " + self.labels[i]
                if i % 2 == 1:
                    lines[-1] += "\n"
            elif self.num_kpts > 0:
                # The following four lines are modified
                if self.labels is not None:
                    lines[-1] += f"{self.kpts_weights[i]:4} {self.labels[i]}"
                else:
                    lines[-1] += f"{self.kpts_weights[i]:4}"

        # Print shifts for automatic kpoints types if not zero.
        if self.num_kpts <= 0 and tuple(self.kpts_shift) != (0, 0, 0):
            lines.append(" ".join([str(x) for x in self.kpts_shift]))
        return "\n".join(lines) + "\n"






# -*- coding: utf-8 -*-
#  Copyright (c) 2021. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from typing import List, Optional, Dict, Any

import numpy as np
from monty.json import MSONable

try:
    from phonopy import Phonopy
    from phonopy.interface.vasp import read_vasp_from_strings, \
        parse_force_constants, get_born_vasprunxml
    from phonopy.structure.atoms import PhonopyAtoms
    from pymatgen.electronic_structure.boltztrap2 import VasprunBSLoader, \
        BztInterpolator, BztTransportProperties
except ImportError:
    raise ImportError('Calculating effective mass requires BoltzTrap2')

from pymatgen.core import IStructure, Structure
from pymatgen.io.vasp import Vasprun
from vise.util.bravais_lattice import BravaisLattice
from vise.util.centering import Centering
from vise.util.logger import get_logger
from vise.util.mix_in import ToJsonFileMixIn
from vise.util.structure_symmetrizer import StructureSymmetrizer


logger = get_logger(__name__)


@dataclass
class PhonopyInput(MSONable, ToJsonFileMixIn):
    """Parameters for non-analytical term correction

    dict
        Parameters used for non-analytical term correction
        'born': ndarray
            Born effective charges
            shape=(primitive cell atoms, 3, 3), dtype='double', order='C'
        'factor': float
            Unit conversion factor
        'dielectric': ndarray
            Dielectric constant tensor
            shape=(3, 3), dtype='double', order='C'

    """
    unitcell: Structure
    supercell: Structure
    supercell_matrix: List[List[float]]
    force_constants: Optional[np.ndarray] = None
    nac_params: Optional[Dict[str, Any]] = None

    @property
    def to_phonopy(self) -> Phonopy:
        if self.force_constants is None:
            raise ValueError("Set force_constants first.")
        _nac = "" if self.nac_params else "*not* "
        logger.info(f"Parameters for a non-analytical term correction are "
                    f"{_nac}set.")

        result = Phonopy(unitcell=structure_to_phonopy_atoms(self.unitcell),
                         supercell_matrix=self.supercell_matrix,
                         nac_params=self.nac_params)
        result.force_constants = self.force_constants
        return result

    def set_force_constants_from_vasprun(self, vasprun_name: str) -> None:
        self.force_constants, _ = parse_force_constants(vasprun_name)
        vasprun = Vasprun(vasprun_name)
        if vasprun.vasp_version[0] == "6":
            self.force_constants /= 15.633302 ** 2

    def set_nac_params_from_vasprun(self, vasprun_name: str) -> None:
        born, epsilon, _, = get_born_vasprunxml(
            vasprun_name, primitive_matrix=np.eye(3),
            supercell_matrix=self.supercell_matrix)
        self.nac_params = {"born": born, "dielectric": epsilon}
        # XX: need factor?


def structure_to_phonopy_atoms(structure: IStructure) -> PhonopyAtoms:
    return read_vasp_from_strings(structure.to(fmt="poscar"), None)


def make_phonopy_input(unitcell: Structure,
                       supercell_matrix: List[List[float]] = None,
                       conventional_base: bool = True) -> PhonopyInput:
    symmetrizer = StructureSymmetrizer(unitcell)
#    assert symmetrizer.is_primitive_lattice_changed
    supercell_matrix = (supercell_matrix
                        or default_supercell_matrix(symmetrizer.bravais))

    if conventional_base:
        centering = Centering(symmetrizer.centering)
        p_to_c_matrix = centering.primitive_to_conv
        supercell_matrix = np.dot(supercell_matrix, p_to_c_matrix)

    supercell = symmetrizer.primitive * supercell_matrix
    return PhonopyInput(unitcell=symmetrizer.primitive, supercell=supercell,
                        supercell_matrix=supercell_matrix.tolist())


def default_supercell_matrix(bravais: BravaisLattice) -> np.ndarray:
    if bravais in [BravaisLattice.hR, BravaisLattice.hP]:
        return np.array([[2, 1, 0], [1, 2, 0], [0, 0, 2]])
    else:
        return np.eye(3) * 2




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest

from vise.input_set.xc import Xc
from vise.tests.helpers.assertion import assert_msonable


def test_xc_str():
    assert "pbe" == str(Xc.pbe)


def test_task_msonable():
    assert_msonable(Xc.pbe)


def test_xc_from_string():
    assert Xc.from_string("pbe") == Xc.pbe
    assert Xc.from_string("perdew-zunger81") == Xc.lda


def test_xc_from_string_raise_error():
    with pytest.raises(AttributeError):
        Xc.from_string("not_exist")


def test_xc_is_lda_or_gga():
    assert Xc.pbe.is_lda_or_gga
    assert Xc.hse.is_lda_or_gga is False


def test_xc_is_hybrid_functional():
    assert Xc.hse.is_hybrid_functional is True


def test_xc_is_local_or_semilocal():
    assert Xc.pbe.is_local_or_semilocal is True
    assert Xc.scan.is_local_or_semilocal is True
    assert Xc.hse.is_local_or_semilocal is False


def test_xc_is_nonlocal():
    assert Xc.scan.is_nonlocal is False
    assert Xc.hse.is_nonlocal is True




# -*- coding: utf-8 -*-
#  Copyright (c) 2022 Kumagai group.
import numpy as np

# unit conversion and constants used in vasp.5.4.4 src
RYTOEV = 13.605826  # Rydberg to eV
AUTOA = 0.529177249  # atomic unit to angstrom
PI = 3.141592653589793238


def vasp_grid(encut: float, lattice_length: float, symprec: str) -> int:
    """ Calculate the numbers of FFT grids used in VASP

    << src from main.F of vasp.5.4.4>>
    XCUTOF =SQRT(INFO%ENMAX /RYTOEV)/(2*PI/(LATT_CUR%ANORM(1)/AUTOA))
    IF (INFO%SZPREC(1:1)=='h' .OR. INFO%SZPREC(1:1)=='a') THEN
      WFACT=4
    ELSE
      WFACT=3
    ENDIF
    GRID%NGPTAR(1)=XCUTOF*WFACT+0.5_q
    CALL FFTCHK(GRID%NGPTAR)
    <<end src>>
    FFTCHK update original grids based on specific algorithms,
     see setup_actual_grids() in this module file for details.
    """
    # In case of ACCURATE or HIGH, factor = 4
    factor = 4 if symprec[0].lower() in ["a", "h"] else 3
    cutoff_in_rydberg = encut / RYTOEV
    lattice_length_in_angstrom = lattice_length / AUTOA
    rec_lat_len = 2 * PI / lattice_length_in_angstrom
    return int(round(factor * np.sqrt(cutoff_in_rydberg) / rec_lat_len))





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from argparse import Namespace
from pathlib import Path

from vise.analyzer.atom_grouping_type import AtomGroupingType
from vise.analyzer.plot_diele_func_data import DieleFuncPlotType
from vise.cli.main import parse_args
from vise.defaults import defaults
from vise.input_set.task import Task
from vise.input_set.xc import Xc

parent_dir = Path(__file__).parent


def test_structure_info_wo_options():
    parsed_args = parse_args(["si"])
    expected = Namespace(
        poscar="POSCAR",
        symprec=defaults.symmetry_length_tolerance,
        angle_tolerance=defaults.symmetry_angle_tolerance,
        show_conventional=False,
        show_primitive=False,
        func=parsed_args.func)
    assert parsed_args == expected


def test_structure_info_w_options():
    parsed_args = parse_args(["si",
                              "-p", "a",
                              "-s", "1",
                              "-a", "2",
                              "-c"])
    expected = Namespace(
        poscar="a",
        symprec=1.0,
        angle_tolerance=2.0,
        show_conventional=True,
        show_primitive=False,
        func=parsed_args.func)
    assert parsed_args == expected


def test_get_poscars_w_mpid():
    parsed_args = parse_args(["gp", "-m", "mp-1234"])
    # func is a pointer so need to point the same address.
    expected = Namespace(
        mpid="mp-1234",
        formula=None,
        func=parsed_args.func)
    assert parsed_args == expected


def test_get_poscars_w_formula():
    parsed_args = parse_args(["get_poscar", "-f", "H"])

    expected = Namespace(
        mpid=None,
        formula="H",
        func=parsed_args.func)
    assert parsed_args == expected


def test_vasp_set_wo_options():
    parsed_args = parse_args(["vs"])
    # func is a pointer so need to point the same address.
    expected = Namespace(
        dirs=[Path.cwd()],
        poscar=None,
        task=defaults.task,
        xc=defaults.xc,
        kpt_density=None,
        overridden_potcar=defaults.overridden_potcar,
        user_incar_settings=None,
        prev_dir=None,
        vasprun=defaults.vasprun,
        outcar=defaults.outcar,
        options=None,
        uniform_kpt_mode=False,
        file_transfer_type=None,
        func=parsed_args.func,
        )
    assert parsed_args == expected


def test_vasp_set_w_options():
    parsed_args = parse_args(["vs",
                              "-d", "Va_O1_0", "Va_O1_1",
                              "--poscar", "POSCAR-tmp",
                              "-t", "band",
                              "-x", "pbesol",
                              "-k", "4.2",
                              "--potcar", "Mg_pv", "O_h",
                              "--user_incar_settings", "LREAD", "F",
                              "-pd", "c",
                              "--vasprun", "vasprun_1",
                              "--outcar", "OUTCAR_1",
                              "--options", "encut", "800",
                              "--uniform_kpt_mode",
                              "--file_transfer", "WAVECAR", "C",
                              ])

    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        poscar=Path("POSCAR-tmp"),
        task=Task.band,
        xc=Xc.pbesol,
        kpt_density=4.2,
        overridden_potcar=["Mg_pv", "O_h"],
        user_incar_settings=["LREAD", "F"],
        prev_dir=Path("c"),
        vasprun=Path("vasprun_1"),
        outcar=Path("OUTCAR_1"),
        options=["encut", "800"],
        uniform_kpt_mode=True,
        file_transfer_type=["WAVECAR", "C"],
        func=parsed_args.func,
    )

    assert parsed_args == expected


def test_plot_band_wo_options():
    parsed_args = parse_args(["pb"])
    # func is a pointer so need to point the same address.
    expected = Namespace(
        vasprun=defaults.vasprun,
        kpoints_filename="KPOINTS",
        y_range=[-10.0, 10.0],
        filename="band.pdf",
        # wavecar_filename=None,
        poscar="POSCAR",
        func=parsed_args.func,
    )
    assert parsed_args == expected


def test_plot_band_w_options():
    parsed_args = parse_args(["pb",
                              "--vasprun", "vasprun_1",
                              "--kpoints", "KPOINTS_1",
                              "--y_range", "-1.0", "1.0",
                              "--filename", "band_1.pdf",
                              # "--wavecar", "WAVECAR",
                              "--poscar", "POSCAR-test"
                              ])

    expected = Namespace(
        vasprun=Path("vasprun_1"),
        kpoints_filename="KPOINTS_1",
        y_range=[-1.0, 1.0],
        filename="band_1.pdf",
        # wavecar_filename="WAVECAR",
        poscar="POSCAR-test",
        func=parsed_args.func,
    )

    assert parsed_args == expected


def test_plot_dos_wo_options():
    parsed_args = parse_args(["pd"])
    # func is a pointer so need to point the same address.
    expected = Namespace(
        vasprun=defaults.vasprun,
        outcar=defaults.outcar,
        type=AtomGroupingType.non_equiv_sites,
        legend=True,
        crop_first_value=True,
        x_range=None,
        y_max_ranges=None,
        target=None,
        filename="dos.pdf",
        base_energy=None,
        title=None,
        func=parsed_args.func,
    )
    assert parsed_args == expected


def test_plot_dos_w_options():
    parsed_args = parse_args(["pd",
                              "--vasprun", "vasprun_1",
                              "--outcar", "OUTCAR_1",
                              "-t", "atoms",
                              "-l", "False",
                              "-c", "False",
                              "--x_range", "-1.0", "1.0",
                              "-y", "-5.0", "5.0",
                              "--target", "1", "2",
                              "--filename", "dos_1.pdf",
                              "-b", "-1",
                              "--title", "test"
                              ])

    expected = Namespace(
        vasprun=Path("vasprun_1"),
        outcar=Path("OUTCAR_1"),
        type=AtomGroupingType.atoms,
        legend=False,
        crop_first_value=False,
        x_range=[-1.0, 1.0],
        y_max_ranges=[-5.0, 5.0],
        target=["1", "2"],
        filename="dos_1.pdf",
        base_energy=-1.0,
        title="test",
        func=parsed_args.func,
    )

    assert parsed_args == expected


def test_plot_diele_func_wo_options():
    parsed_args = parse_args(["pdf"])
    expected = Namespace(
        vasprun=defaults.vasprun,
        outcar=defaults.outcar,
        input_csv_name=None,
        filename=None,
        directions=["ave"],
        y_range=None,
        calc_kk=False,
        ita=0.01,
        plot_type=DieleFuncPlotType.absorption_coeff,
        to_csv=False,
        title=None,
        func=parsed_args.func)
    assert parsed_args == expected


def test_plot_diele_func_w_options():
    parsed_args = parse_args(["pdf",
                              "--vasprun", "vasprun_1",
                              "--outcar", "OUTCAR_1",
                              "--input_csv_name", "diele_func.csv",
                              "-f", "a",
                              "-d", "ave", "xx",
                              "-y", "-5", "5",
                              "-ckk",
                              "-i", "0.1",
                              "--plot_type", "diele_func",
                              "--title", "test title",
                              "--to_csv"])
    expected = Namespace(
        vasprun=Path("vasprun_1"),
        outcar=Path("OUTCAR_1"),
        input_csv_name="diele_func.csv",
        filename="a",
        directions=["ave", "xx"],
        y_range=[-5.0, 5.0],
        calc_kk=True,
        ita=0.1,
        plot_type=DieleFuncPlotType.diele_func,
        to_csv=True,
        title="test title",
        func=parsed_args.func)
    assert parsed_args == expected


def test_band_edge_wo_options():
    parsed_args = parse_args(["be"])
    # func is a pointer so need to point the same address.
    expected = Namespace(
        vasprun=defaults.vasprun,
        outcar=defaults.outcar,
        func=parsed_args.func,
    )
    assert parsed_args == expected


def test_band_edge_w_options():
    parsed_args = parse_args(["be",
                              "--vasprun", "vasprun_1",
                              "--outcar", "OUTCAR_1",
                              ])

    expected = Namespace(
        vasprun=Path("vasprun_1"),
        outcar=Path("OUTCAR_1"),
        func=parsed_args.func,
    )

    assert parsed_args == expected



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from copy import deepcopy

import numpy as np
from pymatgen.io.vasp import Vasprun

from vise.analyzer.effective_mass import EffectiveMass
from pymatgen.electronic_structure.boltztrap import BoltztrapError


def make_effective_mass(vasprun: Vasprun, temp, concentrations, vbm, cbm):
    try:
        from pymatgen.electronic_structure.boltztrap2 import VasprunBSLoader, \
            BztInterpolator, BztTransportProperties
    except BoltztrapError:
        raise ImportError('Calculating effective mass requires BoltzTrap2')

    vasprun = deepcopy(vasprun)
    # Need this to fix the band edge assignment issue in pymatgen.
    vasprun.efermi = (cbm + vbm) / 2
    vl = VasprunBSLoader(vasprun)
    energy_range = (cbm - vbm) / 2 + 2.0
    bi = BztInterpolator(vl, energy_range=energy_range)
    btp = BztTransportProperties(bi, temp_r=np.array([temp]))
    btp.compute_properties_doping(concentrations)
    return EffectiveMass(p=btp.Effective_mass_doping["p"].tolist()[0],
                         n=btp.Effective_mass_doping["n"].tolist()[0],
                         temperature=temp,
                         concentrations=concentrations)


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from typing import Dict, List

from pymatgen.core.structure import Structure

from vise.util.enum import ExtendedEnum
from vise.util.structure_symmetrizer import StructureSymmetrizer


class AtomGroupingType(ExtendedEnum):
    non_equiv_sites = "non_equiv_sites"
    elements = "elements"
    atoms = "atoms"

    def grouped_atom_indices(self, structure, target=None):
        if self is self.atoms:
            grouped_indices = group_by_atoms(structure, target)
        elif self is self.elements:
            grouped_indices = group_by_elements(structure, target)
        else:
            grouped_indices = group_by_non_equiv_sites(structure)

        return grouped_indices


def group_by_atoms(structure: Structure, target):
    result = {}
    for num_str in target:
        num_list = [int(x) for x in num_str.split(",")]
        result[num_str] = num_list
    max_atom_index = max([max(val) for val in result.values()])
    if max_atom_index > len(structure):
        raise ValueError("Atom index is out of range.")
    return result


def group_by_elements(structure: Structure,
                      target: List[str] = None) -> Dict[str, List[int]]:
    result = {}
    target = target or [str(e) for e in structure.composition.elements]
    for elem in target:
        result[elem] = \
            [i for i, site in enumerate(structure) if elem in site]
    return result


def group_by_non_equiv_sites(structure: Structure):
    symmetrizer = StructureSymmetrizer(structure)
    return symmetrizer.grouped_atom_indices()





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np
import pytest
from pymatgen.core import Lattice
from pymatgen.core.structure import Structure
from pymatgen.core import Element
from vise.tests.helpers.assertion import assert_msonable

from vise.util.structure_symmetrizer import (
    cell_to_structure, StructureSymmetrizer, Site, num_symmetry_operation,
    first_structure_is_primitive)
from vise.util.bravais_lattice import BravaisLattice


def test_cell_to_structure():
    lattice = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    coords = [[0.0, 0.0, 0.0]]
    cell = (lattice, coords, [1])

    expected = Structure(lattice=lattice, species="H", coords=coords)
    assert cell_to_structure(cell) == expected


@pytest.fixture()
def symmetrizer_mc():
    lattice = [[10.0,  0.0,  0.0],
               [ 0.0, 10.0,  0.0],
               [-2.0,  0.0, 10.0]]
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.0]]
    structure = Structure(lattice=lattice, species=["H"] * 2, coords=coords)
    symmetrizer = StructureSymmetrizer(structure)
    return symmetrizer


def test_repr(symmetrizer_mc):
    expected = """Symprec: 0.01
Angle tolerance: 5.0
Space group: C2/m
Is primitive: False
site    wyckoff    site sym    equiv sites
------  ---------  ----------  -------------
H1      a          2/m         0 1"""
    assert symmetrizer_mc.__repr__() == expected


def test_structure_symmetrizer_mc_primitive(symmetrizer_mc):
    matrix = symmetrizer_mc.primitive.lattice.matrix
    expected = np.array([[5., -5., 0.],
                         [5., 5., 0.],
                         [-2., 0., 10.]])
    np.testing.assert_array_almost_equal(matrix, expected)


def test_structure_symmetrizer_mc_conventional(symmetrizer_mc):
    matrix = symmetrizer_mc.conventional.lattice.matrix
    expected = np.array([[10.,  0.,  0.],
                         [ 0., 10.,  0.],
                         [-2.,  0., 10.]])
    np.testing.assert_array_almost_equal(matrix, expected)


def test_structure_symmetrizer_mc_band_primitive(symmetrizer_mc):
    band_matrix = symmetrizer_mc.band_primitive.lattice.matrix
    expected = np.array([[5., 5., 0.],
                         [-5., 5., 0.],
                         [-2., 0., 10.]])
    np.testing.assert_array_almost_equal(band_matrix, expected)


def test_structure_symmetrizer_mc_irreducible_kpoints(symmetrizer_mc):
    num_kpt_list = [2, 2, 2]
    kpt_shift = [0.5, 0.5, 0.5]
    actual = symmetrizer_mc.irreducible_kpoints(num_kpt_list, kpt_shift)
    expected = [([0.25, 0.25, 0.25], 4), ([0.75, 0.25, 0.25], 4)]
    assert actual == expected


def test_structure_symmetrizer_mc_lattice_change(symmetrizer_mc):
    assert symmetrizer_mc.is_band_primitive_lattice_changed is True
    assert symmetrizer_mc.is_primitive_lattice_changed is True
    assert symmetrizer_mc.band_primitive_differ_primitive is True


def test_sg_number(symmetrizer_mc):
    assert symmetrizer_mc.sg_number == 12


@pytest.fixture()
def symmetrizer_bcc():
    lattice = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]]
    structure = Structure(lattice=lattice, species=["H", "H"],
                          coords=coords)
    symmetrizer = StructureSymmetrizer(structure)
    return symmetrizer


def test_structure_symmetrizer_bcc_primitive(symmetrizer_bcc):
    matrix = symmetrizer_bcc.primitive.lattice.matrix
    expected = np.array([[-0.5,  0.5,  0.5],
                         [ 0.5, -0.5,  0.5],
                         [ 0.5,  0.5, -0.5]])
    np.testing.assert_array_almost_equal(matrix, expected)


def test_structure_symmetrizer_bcc_lattice_change(symmetrizer_bcc):
    assert symmetrizer_bcc.is_band_primitive_lattice_changed is True
    assert symmetrizer_bcc.is_primitive_lattice_changed is True
    assert symmetrizer_bcc.band_primitive_differ_primitive is False


def test_structure_symmetrizer_bcc_irreducible_kpoints(symmetrizer_bcc):
    num_kpt_list = [2, 2, 2]
    kpt_shift = [0.0, 0.0, 0.0]
    actual = symmetrizer_bcc.irreducible_kpoints(num_kpt_list, kpt_shift)
    expected = [([0.0, 0.0, 0.0], 1), ([0.5, 0.0, 0.0], 3),
                ([0.5, 0.5, 0.0], 3), ([0.5, 0.5, 0.5], 1)]
    assert actual == expected


def test_structure_symmetrizer_sc_irreducible_kpoints():
    lattice = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    coords = [[0.0, 0.0, 0.0]]
    structure = Structure(lattice=lattice, species=["H"], coords=coords)
    symmetrizer = StructureSymmetrizer(structure)

    num_kpt_list = [2, 2, 2]
    kpt_shift = [0.5, 0.5, 0.5]
    actual = symmetrizer.irreducible_kpoints(num_kpt_list, kpt_shift)
    expected = [([0.25, 0.25, 0.25], 8)]
    assert actual == expected


def test_grouped_atom_indices(complex_monoclinic_structure):
    symmetrizer = StructureSymmetrizer(complex_monoclinic_structure)
    actual = symmetrizer.grouped_atom_indices()
    assert actual == {'H1_a': [0], 'He1_m': [1, 2], 'He2_m': [3, 4]}


def test_sites(complex_monoclinic_structure):
    symmetrizer = StructureSymmetrizer(complex_monoclinic_structure)
    actual = symmetrizer.sites
    expected = {"H1": Site("H", "a", "2/m", [0]),
                "He1": Site("He", "m", "m", [1, 2]),
                "He2": Site("He", "m", "m", [3, 4])}
    assert actual == expected


def test_sites_for_bio2():
    bio2_structure = Structure.from_str("""Bi4 O8
    1.0
    4.880207 0.000000 -2.019673
             -0.667412 5.402283 -1.612690
                       -0.040184 -0.016599 6.808846
    Bi O
    4 8
    direct
    0.729106 0.250000 1.000000 Bi
    0.270894 0.750000 0.000000 Bi
    0.000000 0.000000 0.500000 Bi
    0.500000 0.500000 0.500000 Bi
    0.321325 0.021684 0.814642 O
    0.114894 0.650123 0.351314 O
    0.885106 0.349877 0.648686 O
    0.763580 0.849877 0.648686 O
    0.236420 0.150123 0.351314 O
    0.493318 0.521684 0.814642 O
    0.506682 0.478316 0.185358 O
    0.678675 0.978316 0.185358 O""", fmt="POSCAR")
    symmetrizer = StructureSymmetrizer(bio2_structure)
    actual = symmetrizer.sites
    expected = \
        {'Bi1': Site(element='Bi', wyckoff_letter='e', site_symmetry='2', equivalent_atoms=[0, 1]),
         'Bi2': Site(element='Bi', wyckoff_letter='c', site_symmetry='-1', equivalent_atoms=[2, 3]),
         'O1': Site(element='O', wyckoff_letter='f', site_symmetry='1', equivalent_atoms=[4, 9, 10, 11]),
         'O2': Site(element='O', wyckoff_letter='f', site_symmetry='1', equivalent_atoms=[5, 6, 7, 8])}
    assert actual == expected


def test_bravais_lattice():
    lattice =[[10.0,  0.0,  0.0],
              [ 0.0, 10.0,  0.0],
              [-2.0,  0.0, 10.0]]
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.0]]
    structure = Structure(lattice=lattice, species=["H"] * 2, coords=coords)
    symmetrizer = StructureSymmetrizer(structure)
    assert symmetrizer.bravais == BravaisLattice.mC
    assert symmetrizer.centering == "C"


def test_species_order():
    lattice = Lattice.orthorhombic(5, 6, 7)
    coords = [
        [0.0, 0.0, 0.0],
        [0.5, 0.5, 0.0],
        [0.5, 0.0, 0.5],
        [0.0, 0.5, 0.5],

        [0.0, 0.0, 0.5],
        [0.0, 0.5, 0.0],
        [0.5, 0.0, 0.0],
        [0.5, 0.5, 0.5],
    ]
    structure = Structure(lattice=lattice,
                          species=["H"] * 4 + ["He"] * 4,
                          coords=coords)
    supercell = structure * [1, 1, 2]
    actual = [e.specie for e in StructureSymmetrizer(supercell).conventional]
    expected = [Element.H] * 4 + [Element.He] * 4
    assert actual == expected


def test_spglib_cyclic_behavior():
    input_structure = Structure.from_str("""Ca4 Sc2 Sb2 O12
1.0
5.469740 0.000000 0.000000
0.000000 5.632621 0.000000
-5.467914 0.000000 7.830251
O
12
direct
0.344251 0.197893 0.550856 O
0.344251 0.302107 0.050856 O
0.655749 0.802107 0.449144 O
0.655749 0.697893 0.949144 O
0.752694 0.208674 0.946720 O
0.752694 0.291326 0.446720 O
0.247306 0.791326 0.053280 O
0.247306 0.708674 0.553280 O
0.144882 0.034671 0.743977 O
0.144882 0.465329 0.243977 O
0.855118 0.965329 0.256023 O
0.855118 0.534671 0.756023 O
""", fmt="POSCAR")
    ss = StructureSymmetrizer(input_structure)
    sss = StructureSymmetrizer(ss.primitive)
    assert ss.primitive == sss.primitive


@pytest.fixture
def site():
    return Site(element="H",
                wyckoff_letter="a",
                site_symmetry="m3m",
                equivalent_atoms=[0, 1, 2, 3, 5, 6])


def test_site_msonable(site):
    assert_msonable(site)


def test_site_pprint_equiv_atoms(site):
    assert site.pprint_equiv_atoms == "0..3 5 6"


def test_num_symmetry_operation():
    assert num_symmetry_operation("m-3m") == 48
    assert num_symmetry_operation(".m.") == 2


def test_is_first_primitive():
    structure = Structure.from_str("""Hg6
1.0
4.4679886286 2.5795944374 4.4083109087
-4.4679886286 2.5795944374 4.4083109087
0.0000000000 -5.1591888749 4.4083109087
H Hg
1 6 
direct
0.0 0.0 0.0 H
0.2500000000 0.6035253347 0.8964746653 Hg
0.8964746653 0.2500000000 0.6035253347 Hg
0.6035253347 0.8964746653 0.2500000000 Hg
0.7500000000 0.3964746653 0.1035253347 Hg
0.1035253347 0.7500000000 0.3964746653 Hg
0.3964746653 0.1035253347 0.7500000000 Hg""", fmt="poscar")
    structure2 = Structure.from_str("""Hg6
1.0
4.4679886286 2.5795944374 4.4083109087
-4.4679886286 2.5795944374 4.4083109087
0.0000000000 -5.1591888749 4.4083109087
H Hg
1 6 
direct
0.0 0.0 0.0 H
0.750000 0.103525 0.396475 Hg
0.396475 0.750000 0.103525 Hg
0.103525 0.396475 0.750000 Hg
0.250000 0.896475 0.603525 Hg
0.603525 0.250000 0.896475 Hg
0.896475 0.603525 0.250000 Hg
""", fmt="poscar")
    assert first_structure_is_primitive(structure, structure2) is True
    assert first_structure_is_primitive(structure2, structure) is False



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path
from pandas import DataFrame
import pandas as pd
from vise.util.logger import get_logger

logger = get_logger(__name__)

parent_dir = Path(__file__).parent


band_gap = {"GaAs": 1.41,
            "Si": 1.17}


class ExpDieleFunc:
    def __init__(self, material):
        self.material = material
        try:
            self.dataframe = pd.read_csv(parent_dir / f"{material}.csv", sep='\s+')
        except FileNotFoundError:
            logger.warning(f"Csv file for {material} does not exist.")
            raise

#        self.dataframe.absorption = self.dataframe.absorption.astype(float)

    @property
    def energies(self):
        return self.dataframe.energy

    @property
    def dielectric_real(self):
        return self.dataframe.real_part

    @property
    def dielectric_imag(self):
        return self.dataframe.imaginary_part

    @property
    def absorption_coeff(self):
        return self.dataframe.absorption

    @property
    def band_gap(self):
        return band_gap[self.material]

    @property
    def reference(self):
        return {"title": "THE HANDBOOK ON OPTICAL CONSTANTS OF SEMICONDUCTORS",
                "author": "Sadao Adachi",
                "publisher": "World Scientific",
                "doi": "doi.org/10.1142/8480",
                "year": 2012}



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from math import sqrt, pi
from typing import List, Optional

import numpy as np
import pandas as pd
from monty.json import MSONable
from tqdm import tqdm
from vise.util.mix_in import ToJsonFileMixIn, ToCsvFileMixIn
from vise.util.logger import get_logger
from scipy.constants import physical_constants as pc

logger = get_logger(__name__)

eV_to_inv_cm = pc["electron volt-inverse meter relationship"][0] / 100


def absorption_coeff(energy: float, real: float, imag: float) -> float:
    return (2 * sqrt(2) * pi * sqrt(sqrt(real ** 2 + imag ** 2) - real)
            * energy * eV_to_inv_cm)


def refractive_idx_real(e_real: float, e_imag: float) -> float:
    return sqrt(e_real + sqrt(e_real ** 2 + e_imag ** 2)) / sqrt(2)


def refractive_idx_imag(e_real: float, e_imag: float) -> float:
    return sqrt(-e_real + sqrt(e_real ** 2 + e_imag ** 2)) / sqrt(2)


@dataclass
class DieleFuncData(MSONable, ToJsonFileMixIn, ToCsvFileMixIn):
    energies: List[float]  # in eV
    directions: List[str]  # ["xx", "yy", "ave"]
    diele_func_real: List[List[float]]  # [list_of_xx, list_of_xy, list_of_ave]
    diele_func_imag: List[List[float]]  # [list_of_xx, list_of_xy, list_of_ave]
    band_gap: Optional[float] = None  # in eV
    title: Optional[str] = None

    def __post_init__(self):
        try:
            assert len(self.directions) == len(self.diele_func_real)
        except AssertionError:
            print(f"{len(self.directions)} vs {len(self.diele_func_real)}")
            raise

    @property
    def real_columns(self):
        return [f"real_{d}" for d in self.directions]

    @property
    def imag_columns(self):
        return [f"imag_{d}" for d in self.directions]

    @property
    def absorption_columns(self):
        return [f"absorption_{d}" for d in self.directions]

    @property
    def to_dataframe(self) -> pd.DataFrame:
        d = {"energies(eV)": self.energies}
        for x, y in zip(self.real_columns, self.diele_func_real):
            d[x] = y
        for x, y in zip(self.imag_columns, self.diele_func_imag):
            d[x] = y
        for x, y in zip(self.absorption_columns, self.absorption_coeff):
            d[x] = y
        d["band_gap"] = [None] * len(self.energies)
        d["band_gap"][0] = self.band_gap

        return pd.DataFrame.from_dict(d)

    @classmethod
    def from_dataframe(cls, df):
        real_T, imag_T, directions = [], [], []
        show_log = True
        col_items = df.iteritems() if hasattr(df, 'iteritems') else df.items()

        # Processing each column
        for column_name, item in col_items:
            if column_name in ["energies(eV)", "band_gap"]:
                continue
            elif "real" in column_name:
                real_T.append(item.tolist())
                directions.append(column_name.split("_")[-1])
            elif "imag" in column_name:
                imag_T.append(item.tolist())
            elif "absorption" in column_name:
                if show_log:
                    logger.info(f"absorption column is not used for "
                                f"constructing DieleFuncData class instance.")
                    show_log = False
            else:
                raise KeyError("The input CSV does not have proper format.")

        return cls(energies=df["energies(eV)"].tolist(),
                   diele_func_real=real_T,
                   diele_func_imag=imag_T,
                   directions=directions,
                   band_gap=float(df.loc[0, "band_gap"]))

    @property
    def absorption_coeff(self):
        return [[absorption_coeff(e, r, i)
                for e, r, i in zip(self.energies, reals, imags)]
                for reals, imags in
                zip(self.diele_func_real, self.diele_func_imag)]

    @property
    def refractive_idx_real(self):
        return [[refractive_idx_real(r, i) for r, i in zip(reals, imags)]
                for reals, imags in
                zip(self.diele_func_real, self.diele_func_imag)]

    @property
    def refractive_idx_imag(self):
        return [[refractive_idx_imag(r, i) for r, i in zip(reals, imags)]
                for reals, imags in
                zip(self.diele_func_real, self.diele_func_imag)]

    @property
    def reflectivity(self):
        return [[((n - 1) ** 2 + k ** 2) / ((n + 1) ** 2 + k ** 2)
                 for n, k in zip(ns, ks)]
                for ns, ks in zip(self.refractive_idx_real,
                                  self.refractive_idx_imag)]


def min_e_w_target_coeff(energies, quantities, target_value):
    for e, quantity in zip(energies, quantities):
        if quantity > target_value:
            return e
    logger.warning(
        f"The target value {target_value} is not reached in the entire range.")
    return None


def make_shifted_diele_func(diele_func_data: DieleFuncData,
                            original_band_gap: float,
                            shift: float) -> DieleFuncData:
    imag = imag_shift(diele_func_data.diele_func_imag,
                      diele_func_data.energies,
                      original_band_gap + shift, shift)
    real = kramers_kronig_trans(imag, diele_func_data.energies)
    return DieleFuncData(diele_func_data.energies,
                         diele_func_data.directions,
                         real.tolist(),
                         imag.tolist(),
                         original_band_gap + shift)


def imag_shift(diele_func_imag: List[List[float]],
               energies: List[float],
               band_gap: float,
               shift: float) -> np.ndarray:
    energies = np.array(energies)
    assert shift > 0
    result = []
    for energy_grid in energies:
        old_e = energy_grid - shift
        right_idx = np.argwhere(energies > old_e)[0][0]
        left_e, right_e = energies[right_idx - 1], energies[right_idx]
        # linear interpolation
        left_ratio = (right_e - old_e) / (right_e - left_e)

        inner_result = []
        for imag_idx in range(len(diele_func_imag)):
            if energy_grid < band_gap:
                inner_result.append(0.0)
            else:
                old_diele = \
                    diele_func_imag[imag_idx][right_idx - 1] * left_ratio + \
                    diele_func_imag[imag_idx][right_idx] * (1 - left_ratio)
                inner_result.append(
                    old_diele * (energy_grid - shift) / energy_grid)

        result.append(inner_result)

    return np.array(result).T


def kramers_kronig_trans(diele_func_imag: np.array,
                         energies: List[float],
                         ita: float = 0.01) -> np.ndarray:
    mesh = energies[1] - energies[0]
    result = []
    ee2ss = [[e ** 2 - energy_grid ** 2 for e in energies]
             for energy_grid in energies]
    for imag_idx in tqdm(range(len(diele_func_imag))):
        imags = diele_func_imag[imag_idx, :]
        if imag_idx == 0 or \
                (imag_idx > 0
                 and np.allclose(
                            imags, diele_func_imag[imag_idx - 1, :]) is False):
            if np.count_nonzero(imags) == 0:
                inner_result = [0.0] * len(energies)
            else:
                inner_result = []
                for ee2s in ee2ss:
                    integrals = [e * imag * ee2 / (ee2 ** 2 + ita ** 2)
                                 for e, ee2, imag in zip(energies, ee2s, imags)]
                    integral = sum(integrals) * mesh * 2 / pi
                    if imag_idx < 3:
                        integral += 1
                    inner_result.append(integral)

        result.append(inner_result)

    return np.array(result)


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from vise.util.logger import get_logger


def test_get_logger():
    assert str(get_logger("a")) == "<Logger a (DEBUG)>"



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import re
from pathlib import Path
from typing import List

import yaml
from vise.util.logger import get_logger
logger = get_logger(__name__)


class UserSettings:
    def __init__(self, yaml_filename: str):
        self._cwd = Path.cwd()
        self.yaml_filename = yaml_filename
        self.yaml_files_from_root_dir = self._make_yaml_file_list()

    def _make_yaml_file_list(self) -> List[Path]:
        result = []

        dirname = self._cwd
        while True:
            filenames = [self.yaml_filename, "." + self.yaml_filename]
            file_paths = [dirname / filename for filename in filenames]
            for file_path in file_paths:
                if file_path.exists():
                    result.append(file_path)

            if dirname == Path("/"):
                break
            else:
                dirname = dirname.parent

        return list(reversed(result))

    @property
    def user_settings(self) -> dict:
        result = {}

        for file_path in self.yaml_files_from_root_dir:
            logger.info(f"Setting file: {file_path} is parsed...")
            try:
                with open(str(file_path), "r") as fin:
                    settings = yaml.load(fin, Loader=yaml.SafeLoader)
                    for k, v in settings.items():
                        if k in result:
                            logger.info(
                                f"key {k} was overridden to {v} by {file_path}")
                            if isinstance(v, dict):
                                result[k].update(v)
                                continue

                        if self.is_path(v):
                            v = file_path.parent / v
                        result[k] = v
            except AttributeError:
                pass

        logger.info(f"-- Settings from {self.yaml_filename}:")
        logger.info(", ".join(f"{k}: {v} " for k, v in result.items()))
        return result

    @staticmethod
    def is_path(key):
        return isinstance(key, str) and re.match(r'\S*/\S*', key)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np
from pymatgen.io.vasp import Vasprun, Outcar
from vise.analyzer.dielectric_function import DieleFuncData, \
    kramers_kronig_trans
from vise.analyzer.vasp.band_edge_properties import VaspBandEdgeProperties


def make_diele_func(vasprun: Vasprun,
                    outcar: Outcar,
                    use_vasp_real: bool = True,
                    ita: float = 0.01) -> DieleFuncData:

    energies, real, imag = vasprun.dielectric_data["density"]
    imag = np.array(imag)
    if use_vasp_real:
        real = np.array(real)
        real, imag = real.T, imag.T
    else:
        # When CSHIFT = 0.0, the first component becomes 99999.0
        # the following lines are copied from the vasprun.xml
        # <dielectricfunction>
        # <imag>
        # <array>
        # <dimension dim=“1”>gridpoints</dimension>
        # <field>energy</field>
        # <field>xx</field>
        # <field>yy</field>
        # <field>zz</field>
        # <field>xy</field>
        # <field>yz</field>
        # <field>zx</field>
        # <set>
        # <r>     0.0000     0.0000     0.0000    -0.0000 99999.0000 99999.0000 99999.0000 </r>
        imag[0] = 0.0
        imag = imag.T
        real = kramers_kronig_trans(imag, energies, ita)

    real = np.vstack([real, make_average(real)])
    imag = np.vstack([imag, make_average(imag)])
    band_gap = VaspBandEdgeProperties(vasprun, outcar).band_gap
    return DieleFuncData(energies=energies,
                         directions=["xx", "yy", "zz", "xy", "yz", "xz", "ave"],
                         diele_func_real=real.tolist(),
                         diele_func_imag=imag.tolist(),
                         band_gap=band_gap)


def make_average(vals):
    return np.average(vals[:3, :], axis=0)



# -*- coding: utf-8 -*-
#  Copyright (c) 2021. Distributed under the terms of the MIT License.
import re
from copy import copy


def latexify(formula):
    """
    Copied from pymatgen.2021.3.9 as it will be removed from 2022.

    Generates a LaTeX formatted formula. E.g., Fe2O3 is transformed to
    Fe$_{2}$O$_{3}$.

    Args:
        formula (str): Input formula.

    Returns:
        Formula suitable for display as in LaTeX with proper subscripts.
    """
    return re.sub(r"([A-Za-z\(\)])([\d\.]+)", r"\1$_{\2}$", formula)


def numbers_to_lowercases(string: str):
    result = copy(string)
    for i, j in zip(range(10),
                    ["₀", "₁", "₂", "₃", "₄", "₅", "₆", "₇", "₈", "₉"]):
        result = result.replace(str(i), j)
    return result



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from vise.analyzer.dos_data import PDos, DosData, DosBySpinEnergy, \
    scissor_energy, DosPlotData
from vise.analyzer.plot_dos import DosPlotter
from vise.tests.helpers.assertion import assert_msonable, \
    assert_dataclass_almost_equal

energies = [-9, 0, 9]
total = np.array([[0, 5, 0], [0, 5, 0]])


@pytest.fixture()
def orbitals():
    return {"s":   np.array([[0, 0, 0], [1, 0, 0]], dtype=float),
            "px":  np.array([[0, 0, 1], [1, 0, 0]], dtype=float),
            "py":  np.array([[0, 0, 2], [1, 0, 0]], dtype=float),
            "pz":  np.array([[0, 0, 3], [1, 0, 0]], dtype=float),
            "dxy": np.array([[0, 1, 0], [1, 0, 0]], dtype=float),
            "dyz": np.array([[0, 2, 0], [1, 0, 0]], dtype=float),
            "dxz": np.array([[0, 3, 0], [1, 0, 0]], dtype=float),
            "dx2": np.array([[0, 4, 0], [1, 0, 0]], dtype=float),
            "dz2": np.array([[0, 5, 0], [1, 0, 0]], dtype=float),
            "f_3": np.array([[1, 0, 0], [1, 0, 0]], dtype=float),
            "f_2": np.array([[2, 0, 0], [1, 0, 0]], dtype=float),
            "f_1": np.array([[3, 0, 0], [1, 0, 0]], dtype=float),
            "f0":  np.array([[4, 0, 0], [1, 0, 0]], dtype=float),
            "f1":  np.array([[5, 0, 0], [1, 0, 0]], dtype=float),
            "f2":  np.array([[6, 0, 0], [1, 0, 0]], dtype=float),
            "f3":  np.array([[7, 0, 0], [1, 0, 0]], dtype=float)}


@pytest.fixture
def orbitals_2(orbitals):
    return {k: v * 2 for k, v in orbitals.items()}


@pytest.fixture
def orbitals_3(orbitals):
    return {k: v * 3 for k, v in orbitals.items()}


@pytest.fixture
def pdos_list(orbitals, orbitals_2, orbitals_3):
    return [PDos(**orbitals), PDos(**orbitals_2), PDos(**orbitals_3)]


reference_energy = 0.5


@pytest.fixture
def dos_data(pdos_list):
    return DosData(energies=energies, total=total, pdos=pdos_list,
                   base_energy=reference_energy,
                   vertical_lines=[0.0, 1.0])


@pytest.fixture
def dos_data_list(dos_data):
    dos_plot_data_w_lims = dos_data.dos_plot_data(
        grouped_atom_indices={"H": [0], "He": [1, 2]},
        energy_range=[-100, 100],
        dos_ranges=[[-20, 20], [-30, 30], [-40, 40]])
    dos_plot_data_wo_lims = dos_data.dos_plot_data(
        grouped_atom_indices={"H": [0], "He": [1, 2]})

    return dos_data, dos_plot_data_w_lims, dos_plot_data_wo_lims


def test_pdos_msonable(orbitals):
    pdos = PDos(**orbitals)
    assert_msonable(pdos)


def test_dos_data_msonable(dos_data_list):
    dos_data, _, _ = dos_data_list
    assert_msonable(dos_data)


def test_pdos_only_s_p_d_doses(orbitals):
    d = {"s": 1.0, "p": 2.0, "d": 3.0, "f": 4.0}
    pdos = PDos.from_dict(d)
    assert pdos.s == pytest.approx(1.0)
    assert pdos.p == pytest.approx(2.0)
    assert pdos.d == pytest.approx(3.0)
    assert pdos.f == pytest.approx(4.0)


def test_pdos_s_p_d(orbitals):
    p = sum([orbitals[orb] for orb in ["px", "py", "pz"]])
    d = sum([orbitals[orb] for orb in ["dxy", "dyz", "dxz", "dx2", "dz2"]])
    f = sum([orbitals[orb]
             for orb in ["f_3", "f_2", "f_1", "f0", "f1", "f2", "f3"]])
    pdos = PDos(**orbitals)
    assert_array_equal(pdos.p, p)
    assert_array_equal(pdos.d, d)
    assert_array_equal(pdos.f, f)


def test_pdos_wo_f_orbital(orbitals):
    orbitals.pop("f_3")
    assert PDos(**orbitals).f is None


def test_pdos_add(orbitals):
    for pop_orb in ["f_3", "f_2", "f_1", "f0", "f1", "f2", "f3"]:
        orbitals.pop(pop_orb)
    pdos = PDos(**orbitals)
    pdos_sum = pdos + pdos
    assert_array_equal(pdos_sum.s, pdos.s * 2)


def test_dos_data_energies(dos_data_list):
    _, dos_plot_data, _ = dos_data_list
    expected = [e - reference_energy for e in energies]
    assert dos_plot_data.relative_energies == expected


def test_dos_data_lim(dos_data_list):
    _, _, dos_plot_data_wo_lim = dos_data_list
    assert dos_plot_data_wo_lim.energy_range == [-5, 10]
    # (1+..+7)*(2+3)*1.1
    assert dos_plot_data_wo_lim.dos_ranges == \
           [[-5.5, 5.5], [-82.5, 82.5], [-82.5, 82.5]]


def test_dos_plot_data_assertion_error(dos_data):
    with pytest.raises(AssertionError):
        dos_data.dos_plot_data(dos_ranges=[0.1],
                               grouped_atom_indices={"H": [0]})


def test_dos_data_lim_2(dos_data):
    dos_plot_data_w_x_lim = dos_data.dos_plot_data(
        energy_range=[-0.9, -0.1],
        grouped_atom_indices={"H": [0], "He": [1, 2]})
    # (1+2+3+4+5) * (2 + 3)* 1.1 = 82.5
    assert dos_plot_data_w_x_lim.dos_ranges == \
           [[-5.5, 5.5], [-82.5, 82.5], [-82.5, 82.5]]


def test_dos_data_manual_lim(dos_data_list):
    _, dos_plot_data_w_lim, _ = dos_data_list
    assert dos_plot_data_w_lim.energy_range == [-100, 100]
    assert dos_plot_data_w_lim.dos_ranges == [[-20, 20], [-30, 30], [-40, 40]]


def test_dos_data_total_dos(dos_data_list):
    _, dos_plot_data_w_lim, _ = dos_data_list
    assert dos_plot_data_w_lim.doses[0][0].name == ""
    assert_array_equal(dos_plot_data_w_lim.doses[0][0].dos, total)


def test_dos_data_vertical_lines(dos_data_list):
    _, dos_plot_data_w_lim, _ = dos_data_list
    assert dos_plot_data_w_lim.energy_lines == [-0.5, 0.5]


def test_dos_data_pdos_single(pdos_list, dos_data_list):
    _, dos_plot_data_w_lim, _ = dos_data_list
    pdos_sum = pdos_list[1] + pdos_list[2]

    assert dos_plot_data_w_lim.names == ["total", "H", "He"]

    assert dos_plot_data_w_lim.doses[1][0].name == "s"
    assert dos_plot_data_w_lim.doses[1][1].name == "p"
    assert dos_plot_data_w_lim.doses[2][0].name == "s"

    assert_array_equal(dos_plot_data_w_lim.doses[1][0].dos, pdos_list[0].s)
    assert_array_equal(dos_plot_data_w_lim.doses[1][1].dos, pdos_list[0].p)

    assert_array_equal(dos_plot_data_w_lim.doses[2][0].dos, pdos_sum.s)

    plotter = DosPlotter(dos_plot_data_w_lim)
    plotter.construct_plot()
    plotter.plt.show()


def test_orbital_dos():
    total_up = [0, 1]
    total_down = [2, 3]
    total_array = [total_up, total_down]
    orbital_dos = DosBySpinEnergy("total", total_array)
    assert orbital_dos.max_dos() == max([max(total_up), max(total_down)])


def test_max_y_range():
    pdos = [PDos(s=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 px=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 py=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 pz=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 dxy=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 dyz=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 dxz=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 dx2=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 dz2=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 f_3=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 f_2=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 f_1=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 f0=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 f1=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 f2=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 f3=np.array([[0, 10, 20, 30, 40]], dtype=float),
                 )]

    dos_data = DosData(energies=list(range(-2, 3)),
                       total=np.array([[0, 1, 2, 3, 4]]),
                       pdos=pdos,
                       vertical_lines=[0.0])

    dos_plot_data = dos_data.dos_plot_data(
        grouped_atom_indices={"H": [0]},
        energy_range=[-1.1, 1.1])

    assert dos_plot_data.dos_ranges == [[0, 3.3], [0, 231.0]]


def test_dos_by_spin_energy():
    dos = DosBySpinEnergy(name="test",
                          dos=[[-3, -2, -1, 0], [1,  2,  3, 4]])
    mask = [True, False, False, True]
    assert dos.max_dos(mask) == 3.0


def test_dos_plot_data_default_settings():
    d = dict(relative_energies=[0.0, 0.5, 1.0, 1.5],
             doses=[[DosBySpinEnergy(
                 name="test", dos=[[-3, -2, -1, 0], [1,  2, 3, 4]]).as_dict()]],
             names=["total"],
             energy_lines=[0.5])
    actual = DosPlotData.from_dict(d)
    assert actual.dos_ranges == [[-4.4, 4.4]]
    assert actual.energy_range == [-5, 10]


def test_dos_plot_data_csv():
    dos_plot_data = DosPlotData(
        relative_energies=[0.0, 0.5, 1.0],
        doses=[
            [DosBySpinEnergy(name="",
                             dos=[[2.0, 3.0, 4.0], [-3.0, -2.0, -1.0]])],
            [DosBySpinEnergy(name="s",
                             dos=[[12.0, 13.0, 14.0], [-13.0, -12.0, -11.0]])],
        ],
        names=["total", "H"],
        energy_range=[-5, 10],
        dos_ranges=[[-4.4, 4.4]],
        energy_lines=[0.5])
    dos_plot_data.to_csv_file()
    actual = Path("dos_plot_data.csv").read_text()
    expected = """energy(eV),total__up,total__down,H_s_up,H_s_down,energy_lines
0.0,2.0,-3.0,12.0,-13.0,0.5
0.5,3.0,-2.0,13.0,-12.0,
1.0,4.0,-1.0,14.0,-11.0,
"""
    assert actual == expected
    actual = DosPlotData.from_csv_file("dos_plot_data.csv")
    assert_dataclass_almost_equal(actual, dos_plot_data, digit=3,
                                  check_is_subclass=True)


def test_scissor_energy(pdos_list):
    dos_data = DosData(energies=[0.0, 0.5, 1.0],
                       total=np.array([[4, 0, 5], [4, 0, 5]]),
                       pdos=pdos_list,
                       base_energy=0.0,
                       vertical_lines=[0.0, 1.0])
    actual = scissor_energy(dos_data.dos_plot_data(
        grouped_atom_indices={"H": [0]}), energy_shift=1.0)
    assert actual.relative_energies == [0.0, 1.5, 2.0]
    assert actual.doses[0][0] == DosBySpinEnergy(name="",
                                                 dos=[[4, 0, 5], [4, 0, 5]])
    assert actual.energy_lines == [0.0, 2.0]

# test of MSONable exists in test_plot_dos.py.


"""
TODO:
+ Allow to set vbm, cbm, efermi

- + Shift energy zero to manual value
+ Create from FullDosInfo instance with total and atom- and orbital-decomposed pdos
+ Create class to generate grouped_atom_indices and their names

DONE:

"""



# -*- coding: utf-8 -*-

#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest

from vise.input_set.kpoints_mode import KpointsMode


def test_kpoints_mode_from_string():
    assert KpointsMode.from_string("band") == KpointsMode.band


def test_kpoints_mode_from_string_raise_error():
    with pytest.raises(AttributeError):
        KpointsMode.from_string("fail_string")


def test_kpoints_mode_name_list():
    expected = "band, primitive, uniform"
    assert KpointsMode.names_string() == expected
